from typing import List, Union

from omnisdk.omnitron.endpoints import ChannelBatchRequestEndpoint, \
    ChannelIntegrationActionEndpoint, ChannelAttributeConfigEndpoint, \
    ChannelCategoryNodeEndpoint, ChannelCategoryTreeEndpoint, \
    ChannelProductCategoryEndpoint
from omnisdk.omnitron.endpoints import ChannelProductEndpoint, \
    ChannelProductStockEndpoint, \
    ChannelProductPriceEndpoint, ChannelMappedProductEndpoint
from omnisdk.omnitron.models import Product, IntegrationAction, \
    ChannelAttributeConfig, ProductStock, ProductPrice
from requests import HTTPError

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.data import ProductBatchRequestResponseDto
from channel_app.core.utilities import split_list
from channel_app.omnitron.commands.batch_requests import ProcessBatchRequests
from channel_app.omnitron.constants import ContentType, FailedReasonType, \
    BatchRequestStatus, ResponseStatus


class GetInsertedProducts(OmnitronCommandInterface):
    endpoint = ChannelProductEndpoint
    path = "inserts"
    BATCH_SIZE = 100
    content_type = ContentType.product.value

    def get_data(self) -> List[Product]:
        limit = self.BATCH_SIZE
        if self.objects and "limit" in self.objects:
            limit = self.objects["limit"]
        products = self.get_products(limit=limit)
        return products

    @property
    def update_state(self, *args, **kwargs):
        return BatchRequestStatus.commit

    def validated_data(self, data: List[Product]) -> List[Product]:
        formatted_data = [product for product in data
                          if not getattr(product, "failed_reason_type", None)]
        return formatted_data

    def get_products(self, limit: int) -> List[Product]:
        params = {"product_type": 0, "limit": limit}
        language = getattr(self, "param_language", None)
        if language:
            products = self.endpoint(
                path=self.path,
                channel_id=self.integration.channel_id).list(
                headers={"Accept-Language": language}, params=params)
        else:
            products = self.endpoint(
                path=self.path,
                channel_id=self.integration.channel_id).list(params=params)

        products = products[:limit]

        object_list = self.create_batch_objects(
            data=products,
            content_type=ContentType.product.value)
        self.update_batch_request(object_list)

        return products


class GetProductCategoryNodes(OmnitronCommandInterface):
    endpoint = ChannelCategoryNodeEndpoint
    BATCH_SIZE = 100
    content_type = ContentType.product.value

    def get_data(self) -> List[Product]:
        """
        :return:
        """
        products = self.objects
        self.get_product_category(products)
        return products

    def normalize_response(self, data, response) -> List[object]:
        failed_products = [failed_product[0] for failed_product in
                           self.failed_object_list]
        object_list = self.create_batch_objects(
            data=failed_products,
            content_type=ContentType.product.value)
        self.update_batch_request(object_list)
        return data

    def get_product_category(self, products: List[Product]) -> List[Product]:
        if not products:
            empty_list: List[Product] = []
            return empty_list

        category_tree_id = self.integration.channel.category_tree
        category_tree = ChannelCategoryTreeEndpoint(
            channel_id=self.integration.channel_id).retrieve(
            id=category_tree_id)
        category_tree_path = category_tree.category_root["path"]

        product_category_endpoint = ChannelProductCategoryEndpoint(
            channel_id=self.integration.channel_id, path="detailed")

        for product in products:
            product_categories = product_category_endpoint.list(
                params={"product": product.pk})
            for item in product_category_endpoint.iterator:
                product_categories.extend(item)
            category_node_list = []
            for product_category in product_categories:
                if not str(product_category.category["path"]).startswith(
                        category_tree_path):
                    continue
                category_node_list.append(product_category.category)
            if not category_node_list:
                product.failed_reason_type = FailedReasonType.channel_app.value
                self.failed_object_list.append(
                    (product, ContentType.product.value,
                     "ProductCategoryNotFound"))
                continue
            product.category_nodes = category_node_list
        return products


class GetProductCategoryNodesWithIntegrationAction(GetProductCategoryNodes):

    def get_data(self) -> List[Product]:
        """
        :return:
        """
        products = super(GetProductCategoryNodesWithIntegrationAction,
                         self).get_data()
        self.get_category_node_integration_action(products)
        return products

    def get_category_node_integration_action(self, products: List[Product]):
        if not products:
            return []

        category_node_ias_map = {}
        for chunk in split_list(products, 20):
            endpoint = ChannelIntegrationActionEndpoint(
                channel_id=self.integration.channel_id)
            products_category_node_ids = [
                str(product.category_nodes[0].get('pk')) for product
                in chunk if product.category_nodes[0].get('pk')]
            category_node_ias = endpoint.list(
                params={"object_id__in": ",".join(products_category_node_ids),
                        "content_type_name": ContentType.category_node.value,
                        "channel_id": self.integration.channel_id,
                        "sort": "id"
                        })

            for category_node_ias_batch in endpoint.iterator:
                category_node_ias.extend(category_node_ias_batch)

            category_node_integrations_by_id = {ia.object_id: ia for ia in
                                                category_node_ias}
            category_node_ias_map.update(category_node_integrations_by_id)

        for product in products:
            product.category_nodes[0][
                'integration_action'] = category_node_ias_map.get(
                product.category_nodes[0].get('pk'))

        return products


class GetMappedProducts(OmnitronCommandInterface):
    endpoint = ChannelMappedProductEndpoint
    content_type = ContentType.product.value

    def get_data(self) -> List[Product]:
        products = self.get_mapping(self.objects)
        return products

    def normalize_response(self, data, response) -> List[object]:
        failed_products = [failed_product[0] for failed_product in
                           self.failed_object_list]
        object_list = self.create_batch_objects(
            data=failed_products,
            content_type=ContentType.product.value)
        self.update_batch_request(object_list)
        return data

    def validated_data(self, data) -> List[Product]:
        """
        mapped_attributes = {
            "pk": 370,
            "mapped_attributes": {
                "Araba Ağırlığı": "",
                "Yaş Grubu": "",
                "Garanti Süresi": "2",
                "Taşıma Kapasitesi": "",
                "Renk": "Kırmızı",
                "Ek Özellikler": "WHITE_fixed-deger-123_1",
                "Taşma Emniyeti": "WHITE",
                "Cinsiyet": "std",
                "Desi": "",
                "Materyal": "WHITE"
            },
            "attribute_set_id": 1,
            "attribute_set_name": "3 Tekerlekli Bebek Arabası",
            "attribute_set_remote_id": null,
            "mapped_attribute_values": {
                "attribute_omnitron_id": {
                    "attribute_name": "Cinsiyet",
                    "label": "Erkek",
                    "value": "1",
                    "attribute_remote_id": 22,
                    "is_required": false,
                    "is_variant": false,
                    "is_custom": false,
                    "is_meta": false
                },
                "attribute_omnitron_id": {
                    "attribute_name": "Renk",
                    "attribute_remote_id": 24,
                    "is_required": false,
                    "is_variant": false,
                    "is_custom": true,
                    "is_meta": false
                }
            }
        }
        :param data:
        :return:
        """
        attribute_set_ids = {}
        for product in data:
            self.check_product(product=product,
                               attribute_set_ids=attribute_set_ids)

        return data

    def check_product(self, product, attribute_set_ids):
        if not getattr(product, "mapped_attributes", None):
            return
        attribute_set_id = product.mapped_attributes.attribute_set_id
        if not attribute_set_id:
            return

        mapped_attribute_values = getattr(product.mapped_attributes,
                                          'mapped_attribute_values', None)
        if not mapped_attribute_values:
            return

        if attribute_set_id in attribute_set_ids:
            attribute_config_list = attribute_set_ids[attribute_set_id]
        else:
            params = {"attribute_set": attribute_set_id, "limit": 10}
            attribute_config_list = self.get_attribute_config_list(
                params=params)
            attribute_set_ids[attribute_set_id] = attribute_config_list

        for config in attribute_config_list:
            self.update_and_check_product(config=config, product=product)

    def update_and_check_product(self, config: ChannelAttributeConfig,
                                 product: Product):
        mapped_attributes = product.mapped_attributes
        mapped_attribute_values = product.mapped_attributes.mapped_attribute_values
        attribute_id = str(config.attribute["pk"])
        attribute_value_conf = mapped_attribute_values.get(attribute_id)
        if not attribute_value_conf:
            if not product.mapped_attributes.mapped_attribute_values.get(
                    attribute_id, None):
                product.mapped_attributes.mapped_attribute_values[
                    attribute_id] = {
                    "attribute_name": config.attribute.get("name"),
                    "attribute_remote_id": config.attribute_remote_id,
                    "is_required": config.is_required,
                    "is_variant": config.is_variant,
                    "is_custom": config.is_custom,
                    "is_meta": config.is_meta}
            return False
        try:
            self.check_attribute_value_defined(config, mapped_attributes)
            self.check_required(product, config,
                                mapped_attributes.mapped_attributes)

        except Exception as e:
            product.failed_reason_type = FailedReasonType.channel_app.value
            self.failed_object_list.append(
                (product, ContentType.product.value, str(e))
            )
            return False

        product.mapped_attributes.mapped_attribute_values[attribute_id].update(
            {"is_required": config.is_required,
             "is_variant": config.is_variant,
             "is_custom": config.is_custom,
             "is_meta": config.is_meta})
        return True

    def get_attribute_config_list(self, params: dict) -> List[
        ChannelAttributeConfig]:
        config_endpoint = ChannelAttributeConfigEndpoint(path="detailed",
                                                         channel_id=self.integration.channel_id)
        configs_data = config_endpoint.list(params=params)
        for config_batch in config_endpoint.iterator:
            configs_data.extend(config_batch)
        return configs_data

    def check_attribute_value_defined(self, config, mapped_attributes_obj):
        mapped_attributes = mapped_attributes_obj.mapped_attributes
        mapped_attribute_values = mapped_attributes_obj.mapped_attribute_values
        attribute_id = str(config.attribute["pk"])

        mapped_value = mapped_attributes.get(config.attribute["name"])
        attribute_value_conf = mapped_attribute_values.get(attribute_id, None)
        if mapped_value and not config.is_custom and not attribute_value_conf:
            message = f'{config.attribute["name"]} : {mapped_value} was not defined' \
                      f' for {config.attribute_set["name"]} attribute set'
            raise Exception(message)

    def check_required(self, product, config, mapped_attributes):
        attribute_name = config.attribute["name"]
        mapped_value = mapped_attributes.get(attribute_name)
        message = f'Missing Required AttributeValue for Product sku: {product.sku}, ' \
                  f'{self.integration.channel_id} - {mapped_value} : {attribute_name}'

        if config.is_required and attribute_name not in mapped_attributes:
            raise Exception(message)
        if config.is_required and mapped_attributes[attribute_name] in (
                None, ""):
            raise Exception(message)

    def get_mapping(self, products: List[Product]) -> List[Product]:
        """
        Get mapping output of the products according to the schema
        definitions for this channel and product

        :param products: List[Product]
        :return: List[Product]
        """
        language = getattr(self, "param_language", None)
        if language:
            headers = {"Accept-Language": language}
        else:
            headers = {}
        mapped_product_endpoint = self.endpoint(
            channel_id=self.integration.channel_id)

        for index, product in enumerate(list(products)):
            try:
                attributes = mapped_product_endpoint.retrieve(headers=headers,
                                                              id=product.pk)
                product.mapped_attributes = attributes
            except HTTPError as http_err:
                if http_err.response is not None and http_err.response.status_code == 406:
                    try:
                        error_content = http_err.response.json()
                        fail_message = str(error_content.get('error', error_content))
                    except ValueError:
                        fail_message = http_err.response.text
                
                product.mapped_attributes = {}
                product.failed_reason_type = FailedReasonType.mapping.value
                self.failed_object_list.append(
                    (product, ContentType.product.value, fail_message))
                continue

        return products


class GetMappedProductsWithOutCommit(GetMappedProducts):
    def validated_data(self, data) -> List[Product]:
        return data

    def update_batch_request(self, objects_data: list):
        pass


class GetProductPrices(OmnitronCommandInterface):
    endpoint = ChannelProductPriceEndpoint
    CHUNK_SIZE = 50
    content_type = ContentType.product_price.value

    def get_data(self) -> List[Product]:
        products = self.objects

        self.get_product_price(products)
        return products

    def normalize_response(self, data, response) -> List[object]:
        object_list = []
        failed_products = [failed_product[0] for failed_product in
                           self.failed_object_list]
        product_object_list = self.create_batch_objects(
            data=failed_products,
            content_type=ContentType.product.value)
        object_list.extend(product_object_list)

        self.create_integration_actions(data, object_list)
        self.update_batch_request(object_list)
        return data

    def create_integration_actions(self, data, object_list):
        commit_product_prices = [product.productprice for product in data
                                 if not getattr(product, "failed_reason_type",
                                                None)]
        product_price_object_list = self.create_batch_objects(
            data=commit_product_prices,
            content_type=ContentType.product_price.value)
        object_list.extend(product_price_object_list)

    def get_product_price(self, products: List[Product]) -> List[Product]:
        if not products:
            empty_list: List[Product] = []
            return empty_list

        endpoint = self.endpoint(channel_id=self.integration.channel_id)
        product_ids = [str(p.pk) for p in products]
        prices = []
        for chunk in split_list(product_ids, self.CHUNK_SIZE):
            price_batch = self.get_prices(chunk, endpoint)
            if not price_batch:
                break
            prices.extend(price_batch)

        product_prices = {s.product: s for s in prices}

        for index, product in enumerate(products):
            if getattr(product, "failed_reason_type", None):
                continue
            try:
                product.productprice = product_prices[product.pk]
            except KeyError:
                product.failed_reason_type = FailedReasonType.channel_app.value
                self.failed_object_list.append(
                    (product, ContentType.product.value, "PriceNotFound"))
                continue

        return products

    def get_prices(self, chunk, endpoint):
        price_batch = endpoint.list(params={"product__pk__in": ",".join(chunk),
                                            "price_list": self.integration.catalog.price_list})
        for item in endpoint.iterator:
            if not item:
                break
            price_batch.extend(item)
        return price_batch


class GetProductPricesWithOutCommit(GetProductPrices):
    def create_integration_actions(self, data, object_list):
        pass


class GetProductStocks(OmnitronCommandInterface):
    endpoint = ChannelProductStockEndpoint
    content_type = ContentType.product_stock.value
    CHUNK_SIZE = 50

    def get_data(self) -> List[Product]:
        products = self.objects
        self.get_product_stock(products)
        return products

    def normalize_response(self, data, response) -> List[object]:
        object_list = []
        failed_products = [failed_product[0] for failed_product in
                           self.failed_object_list]
        product_object_list = self.create_batch_objects(
            data=failed_products,
            content_type=ContentType.product.value)
        object_list.extend(product_object_list)

        self.create_integration_actions(data, object_list)
        self.update_batch_request(object_list)
        return data

    def create_integration_actions(self, data, object_list):
        commit_product_stocks = [product.productstock for product in data
                                 if not getattr(product, "failed_reason_type",
                                                None)]
        product_stock_object_list = self.create_batch_objects(
            data=commit_product_stocks,
            content_type=ContentType.product_stock.value)
        object_list.extend(product_stock_object_list)

    def get_product_stock(self, products: List[Product]) -> List[Product]:
        if not products:
            empty_list: List[Product] = []
            return empty_list

        product_ids = [str(p.pk) for p in products if
                       not getattr(p, "failed_reason_type", None)]

        endpoint = self.endpoint(channel_id=self.integration.channel_id)
        stocks = []
        for chunk in split_list(product_ids, self.CHUNK_SIZE):
            stock_batch = self.get_stocks(chunk, endpoint)
            if not stock_batch:
                break
            stocks.extend(stock_batch)

        product_stocks = {s.product: s for s in stocks}

        for index, product in enumerate(products):
            if getattr(product, "failed_reason_type", None):
                continue
            try:
                product.productstock = product_stocks[product.pk]
            except KeyError:
                product.failed_reason_type = FailedReasonType.channel_app.value
                self.failed_object_list.append(
                    (product, ContentType.product.value, "StockNotFound"))
                continue

        return products

    def get_stocks(self, chunk, endpoint):
        stock_batch = endpoint.list(params={"product__pk__in": ",".join(chunk),
                                            "stock_list": self.integration.catalog.stock_list})
        for item in endpoint.iterator:
            if not item:
                break
            stock_batch.extend(item)
        return stock_batch


class GetProductStocksWithOutCommit(GetProductStocks):
    def create_integration_actions(self, data, object_list):
        pass


class GetUpdatedProducts(GetInsertedProducts):
    path = "updates"

    def get_data(self) -> List[Product]:
        products = super(GetUpdatedProducts, self).get_data()
        self.get_integration_actions(products)
        return products

    def get_integration_actions(self, products: List[Product]):
        if not products:
            return []

        return_products_as_dict = {}
        for chunk in split_list(products, 20):
            endpoint = ChannelIntegrationActionEndpoint(
                channel_id=self.integration.channel_id)
            product_ids = [str(product.pk) for product in chunk]
            product_ias = endpoint.list(
                params={"object_id__in": ",".join(product_ids),
                        "content_type_name": ContentType.product.value,
                        "channel_id": self.integration.channel_id,
                        "sort": "id"
                        })

            for product_batch in endpoint.iterator:
                product_ias.extend(product_batch)

            product_integrations_by_id = {ia.object_id: ia for ia in
                                          product_ias}
            return_products_as_dict.update(product_integrations_by_id)

        for product in products:
            if product.pk in return_products_as_dict:
                product_ia = return_products_as_dict[product.pk]
                product.integration_action = product_ia

        return products


class GetInsertedOrUpdatedProducts(GetInsertedProducts):
    path = "inserts_or_updates"


class ProcessProductBatchRequests(OmnitronCommandInterface,
                                  ProcessBatchRequests):
    endpoint = ChannelBatchRequestEndpoint
    content_type = ContentType.product.value
    CHUNK_SIZE = 50
    BATCH_SIZE = 100

    def validated_data(self, data: List[ProductBatchRequestResponseDto]):
        for item in data:
            assert isinstance(item, ProductBatchRequestResponseDto)
        return data

    def send(self, validated_data):
        result = self.process_item(validated_data)
        return result

    def check_run(self, is_ok, formatted_data):
        if not is_ok and self.is_batch_request:
            self.integration.batch_request.objects = None
            self.batch_service(self.integration.channel_id).to_fail(
                self.integration.batch_request)
        return False

    @property
    def update_state(self, *args, **kwargs) -> BatchRequestStatus:
        return BatchRequestStatus.done

    def get_channel_items_by_reference_object_ids(self, channel_response,
                                                  model_items_by_content,
                                                  integration_actions):
        channel_items_by_product_id = {}
        for product_id, product in model_items_by_content["product"].items():
            sku = self.get_barcode(obj=product)
            for channel_item in channel_response:
                # TODO: comment
                if channel_item.sku != sku:
                    continue
                remote_item = channel_item
                channel_items_by_product_id[product_id] = remote_item
                break
        return channel_items_by_product_id


class GetDeletedProducts(OmnitronCommandInterface):
    endpoint = ChannelIntegrationActionEndpoint
    content_type = ContentType.integration_action.value
    path = 'deleted'
    BATCH_SIZE = 100

    def get_data(self) -> List[IntegrationAction]:
        products = self.get_deleted_products_ia()

        return products

    def get_deleted_products_ia(self) -> List[IntegrationAction]:
        products_integration_actions = self.endpoint(
            path=self.path,
            channel_id=self.integration.channel_id).list(
            params={"model": ContentType.product.value,
                    "channel_id": self.integration.channel_id,
                    "limit": self.BATCH_SIZE})

        products_integration_actions = products_integration_actions[
                                       :self.BATCH_SIZE]
        return products_integration_actions


class ProcessDeletedProductBatchRequests(ProcessProductBatchRequests):
    """ Manages product deletion process by taking into account the deletion task on channel """
    endpoint = ChannelBatchRequestEndpoint
    content_type = ContentType.batch_request.value
    CHUNK_SIZE = 50
    BATCH_SIZE = 100

    def get_data(self):
        return self.objects

    def validated_data(self, data: List[ProductBatchRequestResponseDto]):
        for item in data:
            assert isinstance(item, ProductBatchRequestResponseDto)
        return data

    def send(self, validated_data):
        result = self.process_item(validated_data)
        return result

    def process_item(self, channel_response):
        endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)

        remote_ids = []
        fail_remote_ids = []
        integration_actions = []
        for remote_item in channel_response:
            if remote_item.status == ResponseStatus.success:
                remote_ids.append(remote_item.remote_id)
            else:
                fail_remote_ids.append(remote_item.remote_id)

        if remote_ids:
            integration_actions = self.get_integration_actions_for_remote_ids(
                remote_ids)

            # successful integration action objects are deleted
            for integration_action in integration_actions:
                if integration_action.content_type.get(
                        "model") in [ContentType.product.value,
                                     ContentType.product_price.value,
                                     ContentType.product_stock.value,
                                     ContentType.product_image.value]:
                    endpoint.delete(id=integration_action.pk)

        # faulty integration action objects are reported
        if fail_remote_ids:
            fail_integration_actions = self.get_integration_actions_for_remote_ids(
                fail_remote_ids)

            for integration_action_obj in fail_integration_actions:
                integration_action_obj.failed_reason_type = \
                    FailedReasonType.channel_app.value

            objects_data = self.create_batch_objects(
                data=fail_integration_actions,
                content_type=ContentType.integration_action.value)
            # done
            self.update_batch_request(objects_data=objects_data)
        return integration_actions

    def get_integration_actions_for_remote_ids(self, remote_ids):
        if not remote_ids:
            return []
        integration_actions_list = []
        for chunk in split_list(remote_ids, 10):
            endpoint = ChannelIntegrationActionEndpoint(
                channel_id=self.integration.channel_id)
            integration_actions = endpoint.list(params={
                "remote_id__in": ",".join(str(r) for r in chunk),
                "channel": self.integration.channel_id,
                "sort": "id"
            })
            for ia_batch in endpoint.iterator:
                if not ia_batch:
                    break
                integration_actions.extend(ia_batch)
            integration_actions_list.extend(integration_actions)
        return [ial for ial in integration_actions_list if ial.remote_id in remote_ids]


class GetProductObjects(OmnitronCommandInterface):
    endpoint = ChannelProductEndpoint
    content_type = ContentType.product.value
    CHUNK_SIZE = 100
    BATCH_SIZE = 100

    def get_data(self):
        """
        :param self.objects: Union[ProductStock, ProductPrice]

        {
            "pk": 8545,
            "product": 76,
            "stock": 10,
            "stock_list": 3,
            "unit_type": "qty",
            "extra_field": {},
            "sold_quantity_unreported": 0,
            "modified_date": "2021-02-16T10:15:18.856000Z"
        }
        """
        return self.objects

    def validated_data(self, data: Union[ProductStock, ProductPrice]):
        product_ids = []
        for obj in data:
            product_ids.append(str(obj.product))
        return product_ids

    def send(self, validated_data):
        result = self.process_item(validated_data)
        for obj in self.objects:
            obj.product = result.get(obj.product)

        return self.objects

    def process_item(self, validated_data):
        return_products_dict = {}
        for chunk in split_list(validated_data, 20):
            endpoint = ChannelProductEndpoint(
                channel_id=self.integration.channel_id,
            )
            products = endpoint.list(
                params={"pk__in": ",".join(c for c in chunk),
                        "sort": "id"})

            for product in endpoint.iterator:
                if not product:
                    break
                products.extend(product)

            products_dict = {product.pk: product for product in products}
            return_products_dict.update(products_dict)

        return return_products_dict


class GetProductsFromBatchrequest(OmnitronCommandInterface):
    """
    It is the command used to fetch products according to bathcrequest.
    """
    endpoint = ChannelProductEndpoint
    BATCH_SIZE = 100
    CHUNK_SIZE = 100
    content_type = ContentType.product.value

    def get_data(self):
        batch_request = self.objects
        product_integration_actions = self.get_integration_actions_from_batchrequest(
            batch_request)
        integration_action_with_product = self.integration.do_action(
            key="get_content_objects_from_integrations",
            objects=product_integration_actions)
        products = self.convert_integration_action_to_product(
            integration_actions=integration_action_with_product)
        return products

    def convert_integration_action_to_product(self, integration_actions):
        products = []
        for integration_action in integration_actions:
            product = getattr(integration_action, "product", None)
            if product:
                setattr(product, "integration_action", integration_action)
                products.append(product)
        return products

    def get_integration_actions_from_batchrequest(self, batch_request):
        """
        Retrieval of integration actions of product type from omnitron
         according to batch request
        :param batch_request:
        :return: ChannelIntegrationAction
        """
        integration_action_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        batch_integration_action = integration_action_endpoint.list(
            params={
                "local_batch_id": batch_request.local_batch_id,
                "status": "processing",
                "content_type_name": self.content_type,
                "limit": self.CHUNK_SIZE,
                "sort": "id"})
        for batch in integration_action_endpoint.iterator:
            if not batch:
                break
            batch_integration_action.extend(batch)
        return batch_integration_action

    def validated_data(self, data: List[Product]) -> List[Product]:
        formatted_data = [product for product in data
                          if not getattr(product, "failed_reason_type", None)]
        return formatted_data



import functools
from collections import defaultdict

from omnisdk.omnitron.endpoints import (ChannelBatchRequestEndpoint,
                                        ChannelIntegrationActionEndpoint,
                                        ChannelProductEndpoint,
                                        ChannelProductPriceEndpoint,
                                        ChannelProductStockEndpoint,
                                        ChannelProductImageEndpoint)

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.utilities import split_list
from channel_app.omnitron.constants import FailedReasonType, ResponseStatus


class GetBatchRequests(OmnitronCommandInterface):
    """
    Fetches BatchRequest entries. For example, we can fetch
    BatchRequests with status 'sent_to_remote'

    There is no state transition in this command.

     :return: List[BatchRequest] as output of do_action
    """
    endpoint = ChannelBatchRequestEndpoint

    def get_data(self):
        """
        
        [{
            "pk": 22,
            "channel": 6,
            "local_batch_id": "89bb31b0-6700-4a9d-8bae-308ac938649c",
            "remote_batch_id": "028b8e66-2a4c-45b5-912f-9ab90036c78a",
            "content_type": "product",
            "status": "sent_to_remote"
        },]
        :return:
        """
        params = getattr(self, "param_{}".format("params"), None)
        if not params:
            return []

        params.update({"channel": self.integration.channel_id})

        batch_requests = self.endpoint(
            channel_id=self.integration.channel_id
        ).list(params=params)
        return batch_requests


class BatchRequestUpdate(OmnitronCommandInterface):
    endpoint = ChannelBatchRequestEndpoint

    def send(self, validated_data) -> object:
        result = self.endpoint(channel_id=self.integration.channel_id).update(
            id=validated_data.pk, item=validated_data)
        return result


class ProcessBatchRequests(object):
    def get_integration_actions_to_processing(self):
        integration_action_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        batch_integration_action = integration_action_endpoint.list(
            params={
                "local_batch_id": self.integration.batch_request.local_batch_id,
                "channel_id": self.integration.channel_id,
                "status": "processing",
                "limit": self.CHUNK_SIZE,
                "sort": "id"})
        for batch in integration_action_endpoint.iterator:
            if not batch:
                break
            batch_integration_action.extend(batch)
        return batch_integration_action

    def group_model_items_by_content_type(self, items_by_content):
        """
        :return {
            "product": {pk1: product1, pk2: product2, pk3: product3},
            "productstock": {pk1: productstock1, pk2: productstock1, pk3: productstock3},
            "productprice": {pk1: productprice1, pk2: productprice2, pk3: productprice3},
            ...
        }
        """
        batch_items = {}
        for model, model_pks in items_by_content.items():
            if model == "product":
                group_items = self.get_products(model_pks)
            elif model == "productstock":
                group_items = self.get_stocks(model_pks)
            elif model == "productprice":
                group_items = self.get_prices(model_pks)
            elif model == "productimage":
                group_items = self.get_images(model_pks)
            else:
                raise NotImplementedError
            batch_items[model] = group_items
        return batch_items

    def group_integration_actions_by_content_type(self,
                                                  batch_integration_actions):
        """

        :param batch_integration_actions:
        :return: {
            "product": [product_pk1, product_pk2, product_pk3, ...],
            "productprice": [productprice_pk1, productprice_pk2, ...]
        }
        """
        items_by_content = {}
        for integration_action in batch_integration_actions:
            content_type = integration_action.content_type["model"]
            if content_type not in items_by_content:
                items_by_content[content_type] = []
            items_by_content[content_type].append(
                str(integration_action.object_id))
        return items_by_content

    def get_products(self, id_list) -> dict:
        end_point = ChannelProductEndpoint(
            channel_id=self.integration.channel_id)
        products = []
        for chunk_id_list in split_list(id_list, self.CHUNK_SIZE):
            products_batch = end_point.list(
                params={"pk__in": ",".join(chunk_id_list),
                        "limit": len(chunk_id_list)})
            products.extend(products_batch)

        return {s.pk: s for s in products}

    def get_prices(self, id_list: list) -> dict:
        """
        param id_list: productprice pk list
        :return: dict of prices with key product id
        {
            1111:{"pk":2222, "stock":5, ...},
        }
        """
        if not id_list:
            return {}

        endpoint = ChannelProductPriceEndpoint(
            channel_id=self.integration.channel_id)
        prices = []
        for chunk in split_list(id_list, self.CHUNK_SIZE):
            # TODO should we check the size of chunk (len(chunk) == len(stock_batch))
            #  to validate something is missing on omnitron side?
            price_batch = endpoint.list(params={"pk__in": ",".join(chunk),
                                                "limit": len(chunk)})
            prices.extend(price_batch)
        return {p.product: p for p in prices if str(p.pk) in id_list}

    def get_stocks(self, id_list: list) -> dict:
        """
        :param id_list:
        :return: dict of stocks with key product id
        {
            1111:{"pk":3333, price:19.99, ...},
        }
        """
        if not id_list:
            return {}
        endpoint = ChannelProductStockEndpoint(
            channel_id=self.integration.channel_id)
        stocks = []
        for chunk in split_list(id_list, self.CHUNK_SIZE):
            # TODO should we check the size of chunk (len(chunk) == len(stock_batch))
            #  to validate something is missing on omnitron side?
            stock_batch = endpoint.list(params={"pk__in": ",".join(chunk),
                                                "limit": len(chunk)})
            stocks.extend(stock_batch)
            if not stock_batch:
                break
        return {s.product: s for s in stocks if str(s.pk) in id_list}

    def get_images(self, id_list):
        if not id_list:
            return {}
        endpoint = ChannelProductImageEndpoint(
            channel_id=self.integration.channel_id)
        images = []
        for chunk in split_list(id_list, self.CHUNK_SIZE):
            image_batch = endpoint.list(params={"id__in": ",".join(chunk),
                                                "limit": len(chunk)})
            images.extend(image_batch)
            if not image_batch:
                break
        product_images = defaultdict(list)
        [product_images[i.product].append(i) for i in images]
        return product_images

    def get_barcode(self, obj):
        """
        # The barcode uniquely identifying a
        # certain product line across
        remote_id_attribute = ["attributes","barcode"], ["sku"]
        :param obj:
        :return: String (required)
        """
        remote_id_attribute = self.integration.channel.conf.get(
            "remote_id_attribute")
        if remote_id_attribute:
            return self.get_reduce_data(remote_id_attribute, obj.__dict__)
        return obj.sku

    @staticmethod
    def get_reduce_data(code, value):
        try:
            data = functools.reduce(
                lambda d, key: d.get(key, None) if isinstance(d,
                                                              dict) else None,
                code.split("__"), value)
            return data
        except TypeError:
            return None

    def _update_batch_request(self, model_items_by_content):
        object_list = []
        for key, model_items in model_items_by_content.items():
            if isinstance(model_items, dict):
                model_items_obj = [value for key, value in model_items.items()]
            else:
                model_items_obj = []
            objects = self.create_batch_objects(data=model_items_obj,
                                                content_type=key)
            object_list.extend(objects)
        self.update_batch_request(objects_data=object_list)

    def update_other_objects(self, channel_items_by_object_id: dict,
                             model_items_by_content: dict):
        for key, model_items in model_items_by_content.items():
            for reference_object_id, model_item in model_items.items():
                if not isinstance(model_item, list):
                    model_item = [model_item]
                for mi in model_item:
                    try:
                        remote_item = channel_items_by_object_id[
                            reference_object_id]
                    except KeyError:
                        mi.failed_reason_type = FailedReasonType.channel_app.value
                        message = f'This item information was not sent from the channel. ID is {reference_object_id}'
                        self.failed_object_list.append((mi, key, message))
                        continue

                    if remote_item.status == ResponseStatus.fail:
                        mi.failed_reason_type = FailedReasonType.channel_app.value
                        self.failed_object_list.append(
                            (mi, key, remote_item.message))
                    else:
                        mi.remote_id = remote_item.remote_id

    def get_channel_items_by_reference_object_ids(self, channel_response,
                                                  model_items_by_content,
                                                  integration_actions):
        raise NotImplementedError

    def process_item(self, channel_response):
        # [1] Get all integration actions of this batch request
        batch_integration_actions = self.get_integration_actions_to_processing()
        if not batch_integration_actions:
            raise Exception("No records was found not with BatchRequest")

        # [2] Group integration actions by content type and each object_ids of them
        integration_items_by_content = self.group_integration_actions_by_content_type(
            batch_integration_actions)
        # [3] Group model items by content type and their object id
        model_items_by_content = self.group_model_items_by_content_type(
            integration_items_by_content)

        # [4] Link Omnitron and Channel items
        channel_items_by_product_id = self.get_channel_items_by_reference_object_ids(
            channel_response=channel_response,
            model_items_by_content=model_items_by_content,
            integration_actions=batch_integration_actions)

        # [5] Updates statuses of related models by monkey patching them
        # Creates failed_object_list
        self.update_other_objects(channel_items_by_product_id,
                                  model_items_by_content)

        # [6] update batch request and object list
        self._update_batch_request(model_items_by_content)



from unittest.mock import MagicMock, patch
from requests import HTTPError
from omnisdk.base_client import BaseClient
from omnisdk.omnitron.endpoints import (
    ChannelBatchRequestEndpoint,
    ChannelCategoryTreeEndpoint,
    ChannelIntegrationActionEndpoint,
    ChannelProductCategoryEndpoint,
    ChannelProductEndpoint,
    ChannelProductImageEndpoint,
    ChannelProductPriceEndpoint,
    ChannelProductStockEndpoint, 
    ChannelExtraProductStockEndpoint, 
    ChannelExtraProductPriceEndpoint,
    ChannelMappedProductEndpoint,
)
from omnisdk.omnitron.models import (
    ChannelAttributeConfig, 
    ProductStock, 
    ProductPrice,
    ProductImage,
)

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.data import ProductBatchRequestResponseDto
from channel_app.core.tests import BaseTestCaseMixin
from channel_app.omnitron.commands.batch_requests import GetBatchRequests
from channel_app.omnitron.commands.product_images import (
    GetInsertedProductImages, 
    GetUpdatedProductImages,
)
from channel_app.omnitron.commands.product_prices import (
    GetInsertedProductPrices,
    GetInsertedProductPricesFromExtraPriceList, 
    GetProductStocksFromProductPrices, 
    GetUpdatedProductPrices,
    GetUpdatedProductPricesFromExtraPriceList
)
from channel_app.omnitron.commands.product_stocks import (
    GetInsertedProductStocks,
    GetProductPricesFromProductStocks,
    GetUpdatedProductStocks,
    GetUpdatedProductStocksFromExtraStockList, 
    GetInsertedProductStocksFromExtraStockList
)
from channel_app.omnitron.commands.products import (
    GetDeletedProducts,
    GetInsertedProducts,
    GetUpdatedProducts,
    ProcessDeletedProductBatchRequests,
    ProcessProductBatchRequests,
    Product,
    GetMappedProducts,
    GetProductPrices,
    GetProductStocks,
    GetProductCategoryNodes,
    GetProductCategoryNodesWithIntegrationAction,
)
from channel_app.omnitron.constants import (
    BatchRequestStatus,
    ContentType,
    FailedReasonType,
    ResponseStatus
)


class TestGetInsertedProducts(BaseTestCaseMixin):
    """
    Test case for GetInsertedProducts
    
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedProducts
    """

    def setUp(self) -> None:
        self.get_inserted_products = GetInsertedProducts(
            integration=self.mock_integration
        )
        self.sample_products = [
            Product(name='test', failed_reason_type=None),
            Product(name='test2', failed_reason_type='error')
        ]
        self.limit = 1

    @patch.object(GetInsertedProducts, 'get_products')
    def test_get_data(self, mock_get_products):
        mock_get_products.return_value = [self.sample_products[0]]
        self.get_inserted_products.BATCH_SIZE = self.limit

        products = self.get_inserted_products.get_data()

        self.assertEqual(len(products), 1)
        self.assertEqual(products[0].name, 'test')
        mock_get_products.assert_called_once_with(limit=self.limit)

    @patch.object(GetInsertedProducts, 'get_products')
    def test_get_data_with_limit_in_objects_dict(self, mock_get_products):
        mock_get_products.return_value = [self.sample_products[0]]
        self.get_inserted_products.objects = {'limit': self.limit}

        products = self.get_inserted_products.get_data()

        self.assertEqual(len(products), 1)
        self.assertEqual(products[0].name, 'test')
        mock_get_products.assert_called_once_with(limit=self.limit)

    def test_update_state_property(self):
        self.assertEqual(
            self.get_inserted_products.update_state,
            BatchRequestStatus.commit
        )

    def test_validated_data(self):
        validated_products = self.get_inserted_products.validated_data(
            self.sample_products
        )

        self.assertIn(self.sample_products[0], validated_products)
        self.assertNotIn(self.sample_products[1], validated_products)

    @patch('channel_app.core.clients.OmnitronApiClient')
    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetInsertedProducts, 'update_batch_request')
    @patch.object(GetInsertedProducts, 'create_batch_objects')
    @patch.object(ChannelProductEndpoint, 'list')
    def test_get_products(
        self,
        mock_list,
        mock_create_batch_objects,
        mock_update_batch_request,
        mock_get_instance,
        mock_omnitron_api_client,
    ):
        mock_list.return_value = self.sample_products

        products = self.get_inserted_products.get_products(
            limit=self.limit + 1
        )

        self.assertEqual(len(products), 2)
        self.assertEqual(products[0].name, 'test')
        self.assertEqual(products[1].name, 'test2')


class TestGetUpdatedProducts(BaseTestCaseMixin):
    """
    Test case for GetUpdatedProducts
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetUpdatedProducts
    """

    def setUp(self):
        self.get_inserted_products = GetInsertedProducts(
            integration=self.mock_integration
        )
        self.get_updated_products = GetUpdatedProducts(
            integration=self.mock_integration
        )
        self.sample_products = [
            Product(
                pk=1,
                name='test',
                failed_reason_type=None,
                modified_date='2021-01-01T00:00:00Z',
                integration_action=MagicMock()
            ),
            Product(
                pk=2,
                name='test2',
                failed_reason_type='error',
                modified_date='2021-01-01T00:00:00Z',
                integration_action=MagicMock()
            )
        ]

    @patch.object(GetInsertedProducts, 'get_products')
    @patch.object(GetUpdatedProducts, 'get_integration_actions')
    def test_get_data(self, mock_get_integration_actions, mock_get_products):
        mock_get_integration_actions.return_value = self.sample_products
        mock_get_products.return_value = [self.sample_products[0]]
        products = self.get_updated_products.get_data()

        self.assertEqual(len(products), 1)
        self.assertEqual(products[0].name, 'test')

    @patch('channel_app.core.clients.OmnitronApiClient')
    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, 'list')
    def test_get_integration_actions(
        self,
        mock_list,
        mock_get_instance,
        mock_omnitron_api_client
    ):
        example_response = MagicMock()
        example_response.json.return_value = [
            {
                'id': 1,
                'channel': 1,
                'content_type': 'product',
                'remote_id': 1,
                'object_id': 1,
            },
            {
                'id': 2,
                'channel': 1,
                'content_type': 'product',
                'remote_id': 2,
                'object_id': 2,
            }
        ]

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response,
        ):
            products = self.get_updated_products.get_integration_actions(
                self.sample_products
            )

        self.assertEqual(len(products), 2)

    def test_get_integration_actions_without_product(self):
        products = self.get_updated_products.get_integration_actions([])
        self.assertEqual(products, [])


class TestGetInsertedOrUpdatedProducts(
    TestGetInsertedProducts,
    BaseTestCaseMixin
):
    """
    Test case for GetInsertedOrUpdatedProducts
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedOrUpdatedProducts
    """

    def setUp(self) -> None:
        self.get_inserted_products = GetInsertedProducts(
            integration=self.mock_integration
        )
        self.get_inserted_products.path = "inserts_or_updates"
        return super().setUp()


class TestGetDeletedProducts(BaseTestCaseMixin):
    """
    Test case for GetDeletedProducts
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetDeletedProducts
    """

    def setUp(self) -> None:
        self.get_deleted_products = GetDeletedProducts(
            integration=self.mock_integration
        )
        self.products = [
            {
                "pk": 23,
                "channel": 3,
                "content_type": {
                    "id": 1,
                    "app_label": "products",
                    "model": "product"
                },
                "object_id": 1,
                "remote_id": None,
                "version_date": "2023-11-07T08:58:16.079727Z",
                "state": {},
                "modified_date": "2023-11-08T09:11:56.919937Z",
                "local_batch_id": "7c43e5fb-32be-4a18-a6fa-539c9d2485ee",
                "status": "processing",
                "created_date": "2023-11-08T09:11:56.919929Z"
            }
        ]

    @patch.object(GetDeletedProducts, 'get_deleted_products_ia')
    def test_get_data(self, mock_get_deleted_products_ia):
        mock_get_deleted_products_ia.return_value = self.products
        products = self.get_deleted_products.get_data()
        self.assertEqual(len(products), 1)

        product = products[0]
        self.assertEqual(product.get('pk'), 23)

    @patch('channel_app.core.clients.OmnitronApiClient')
    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, '_list')
    def test_get_deleted_products_ia(
        self,
        mock_list,
        mock_get_instance,
        mock_omnitron_api_client
    ):
        example_response = MagicMock()
        example_response.json.return_value = self.products
        mock_list.return_value = example_response
        products_ia = self.get_deleted_products.get_deleted_products_ia()
        self.assertEqual(len(products_ia), 1)

        product = products_ia[0].get_parameters()
        self.assertEqual(product.get('pk'), 23)


class TestGetMappedProducts(BaseTestCaseMixin):
    """
    Test case for GetMappedProducts
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetMappedProducts
    """

    def setUp(self) -> None:
        self.get_mapped_products = GetMappedProducts(
            integration=self.mock_integration
        )
        self.sample_products = [
            Product(
                pk=1,
                name='test',
                failed_reason_type=None,
                modified_date='2021-01-01T00:00:00Z'
            ),
            Product(
                pk=2,
                name='test2',
                failed_reason_type='error',
                modified_date='2021-01-01T00:00:00Z'
            )
        ]

    @patch.object(GetMappedProducts, 'get_mapping')
    def test_get_data(self, mock_get_mapping):
        mock_get_mapping.return_value = self.sample_products
        result = self.get_mapped_products.get_data()
        self.assertEqual(len(result), 2)

    @patch.object(GetMappedProducts, 'check_product')
    def test_validated_data(self, mock_check_product):
        data = self.sample_products
        self.get_mapped_products.validated_data(data)
        self.assertEqual(mock_check_product.call_count, 2)

    @patch.object(GetMappedProducts, 'get_attribute_config_list')
    @patch.object(GetMappedProducts, 'update_and_check_product')
    def test_check_product_gets_attribute_config_list(
        self,
        mock_update_and_check_product,
        mock_get_attribute_config_list
    ):
        product = Product()
        product.mapped_attributes = MagicMock()
        product.mapped_attributes.attribute_set_id = 1
        product.mapped_attributes.mapped_attribute_values = {
            "1": {
                "value": "test"
            }
        }
        mock_get_attribute_config_list.return_value = [
            ChannelAttributeConfig()
        ]
        self.get_mapped_products.check_product(
            product,
            {}
        )
        mock_get_attribute_config_list.assert_called_once()

    @patch.object(GetMappedProducts, 'check_attribute_value_defined')
    @patch.object(GetMappedProducts, 'check_required')
    def test_update_and_check_product_checks_attribute_value_and_required(
        self,
        mock_check_required,
        mock_check_attribute_value_defined
    ):
        product = Product()
        product.mapped_attributes = MagicMock()
        product.mapped_attributes.mapped_attribute_values = {}
        config = ChannelAttributeConfig()
        config.attribute_remote_id = 1
        config.is_required = True
        config.is_variant = True
        config.is_custom = True
        config.is_meta = True
        config.attribute = {
            "pk": 1,
            "name": "test",
        }
        result = self.get_mapped_products.update_and_check_product(
            config,
            product
        )
        self.assertFalse(result)

    @patch.object(GetMappedProducts, 'check_attribute_value_defined')
    @patch.object(GetMappedProducts, 'check_required')
    def test_update_and_check_product(
        self,
        mock_check_required,
        mock_check_attribute_value_defined
    ):
        product = Product()
        product.mapped_attributes = MagicMock()
        product.mapped_attributes.mapped_attribute_values = {
            "1": {
                "value": "test"
            }
        }
        config = ChannelAttributeConfig()
        config.attribute_remote_id = 1
        config.is_required = True
        config.is_variant = True
        config.is_custom = True
        config.is_meta = True
        config.attribute = {
            "pk": 1,
            "name": "test",
        }
        result = self.get_mapped_products.update_and_check_product(
            config,
            product
        )
        self.assertTrue(result)

    @patch.object(GetMappedProducts, 'get_attribute_config_list')
    def test_get_attribute_config_list_returns_configs_data(
        self,
        mock_get_attribute_config_list
    ):
        config = ChannelAttributeConfig()
        mock_get_attribute_config_list.return_value = [
            config
        ]
        result = self.get_mapped_products.get_attribute_config_list(
            {"attribute_set": 1, "limit": 10}
        )
        self.assertEqual(result, [config])

    def test_check_attribute_value_defined_raises_exception_when_mapped_value_not_defined(self):
        mapped_attributes_obj = MagicMock()
        mapped_attributes_obj.mapped_attributes = {"name": "value"}
        mapped_attributes_obj.mapped_attribute_values = {}
        config = ChannelAttributeConfig()
        config.attribute = {"pk": 1, "name": "name"}
        config.attribute_set = {"pk": 1, "name": "name"}
        config.is_custom = False
        with self.assertRaises(Exception):
            self.get_mapped_products.check_attribute_value_defined(
                config,
                mapped_attributes_obj
            )

    def test_check_required_raises_exception_when_required_attribute_missing(self):
        product = Product()
        product.sku = "sku"
        self.get_mapped_products.integration = MagicMock()
        self.get_mapped_products.integration.channel_id = 1
        mapped_attributes = {}
        config = ChannelAttributeConfig()
        config.attribute = {"name": "name"}
        config.is_required = True
        with self.assertRaises(Exception):
            self.get_mapped_products.check_required(
                product,
                config,
                mapped_attributes
            )

    @patch.object(GetMappedProducts, 'get_mapping')
    def test_get_mapping_returns_mapped_products(self, mock_get_mapping):
        product = Product()
        mock_get_mapping.return_value = [product]
        result = self.get_mapped_products.get_mapping([product])
        self.assertEqual(result, [product])

    @patch.object(GetMappedProducts, 'get_mapping')
    def test_get_mapping_returns_empty_list_when_no_products(self, mock_get_mapping):
        mock_get_mapping.return_value = []
        result = self.get_mapped_products.get_mapping([])
        self.assertEqual(result, [])
    
    @patch.object(BaseClient, 'get_instance')
    def test_get_mapping_raise_http_error(
        self,
        mock_get_instance
    ):
        mock_response = MagicMock()
        mock_response.status_code = 406
        mock_response.json.return_value = {"error": "Test error"}

        http_error = HTTPError(response=mock_response)
        mock_endpoint = MagicMock()
        mock_endpoint.retrieve.side_effect = http_error

        with patch.object(
            ChannelMappedProductEndpoint,
            '__new__',
            return_value=mock_endpoint
        ):
            test_product = Product(pk=1)
            products = self.get_mapped_products.get_mapping([test_product])

            self.assertEqual(products[0].mapped_attributes, {})
            self.assertEqual(
                products[0].failed_reason_type,
                FailedReasonType.mapping.value
            )
            
            self.assertEqual(len(self.get_mapped_products.failed_object_list), 1)
            self.assertEqual(
                self.get_mapped_products.failed_object_list[0][0],
                test_product
            )
            self.assertEqual(
                self.get_mapped_products.failed_object_list[0][1], 
                ContentType.product.value
            )
            self.assertEqual(
                self.get_mapped_products.failed_object_list[0][2],
                "Test error"
            )


class TestGetMappedProductsWithOutCommit(TestGetMappedProducts):
    pass


class TestGetProductPrices(BaseTestCaseMixin):
    """
    Test case for GetProductPrices
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetProductPrices
    """

    def setUp(self) -> None:
        self.get_product_prices = GetProductPrices(
            integration=self.mock_integration
        )

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductPrices, 'get_prices')
    def test_successful_product_price_retrieval(
        self,
        mock_get_instance,
        mock_get_prices
    ):
        products = [Product(pk=i, productprice=10) for i in range(1, 6)]
        self.get_product_prices.objects = products

        result = self.get_product_prices.get_data()
        self.assertEqual(result, products)

    @patch.object(GetProductPrices, 'get_prices')
    @patch.object(BaseClient, 'get_instance')
    def test_product_price_retrieval_with_successful_get_product_price(
        self,
        mock_get_instance,
        mock_get_prices
    ):
        products = [Product(pk=i) for i in range(1, 6)]

        price_list = []
        for product in products:
            price = MagicMock()
            price.product = product.pk
            price_list.append(price)

        mock_get_prices.return_value = price_list
        result = self.get_product_prices.get_product_price(products)
        self.assertEqual(result, products)

    @patch.object(GetProductPrices, 'get_prices')
    @patch.object(BaseClient, 'get_instance')
    def test_product_price_retrieval_with_failed_get_product_price(
        self,
        mock_get_instance,
        mock_get_prices
    ):
        products = [Product(pk=i, product_price=0) for i in range(1, 6)]

        price_list = []
        for product in products:
            price = MagicMock()
            price.product = product.pk

            if len(price_list) < len(products) - 1:
                price_list.append(price)

        mock_get_prices.return_value = price_list
        result = self.get_product_prices.get_product_price(products)
        self.assertFalse(hasattr(result[-1], 'productprice'))


class TestGetProductPricesWithOutCommit(TestGetProductPrices):
    pass


class TestGetProductStocks(BaseTestCaseMixin):
    """
    Test case for GetProductStocks
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetProductStocks
    """

    def setUp(self) -> None:
        self.get_product_stocks = GetProductStocks(
            integration=self.mock_integration
        )
        self.sample_products = [
            Product(
                pk=1,
                name='test',
                failed_reason_type=None,
                productstock=10,
                modified_date='2021-01-01T00:00:00Z'
            ),
            Product(
                pk=2,
                name='test2',
                failed_reason_type='error',
                productstock=15,
                modified_date='2021-01-01T00:00:00Z'
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductStocks, 'get_stocks')
    def test_get_data(
        self,
        mock_get_stocks,
        mock_get_instance
    ):
        mock_get_stocks.return_value = []
        self.get_product_stocks.objects = self.sample_products
        result = self.get_product_stocks.get_data()
        self.assertEqual(len(result), 2)

    @patch.object(GetProductStocks, 'create_batch_objects')
    @patch.object(GetProductStocks, 'create_integration_actions')
    @patch.object(GetProductStocks, 'update_batch_request')
    def test_normalize_response(
        self,
        mock_update_batch_request,
        mock_create_integration_actions,
        mock_create_batch_objects
    ):
        data = self.sample_products
        response = MagicMock()
        response.json.return_value = []
        self.get_product_stocks.failed_object_list = []
        result = self.get_product_stocks.normalize_response(data, response)
        self.assertEqual(result, data)

    @patch.object(GetProductStocks, 'create_batch_objects')
    def test_create_integration_actions(self, mock_create_batch_objects):
        data = self.sample_products
        object_list = []
        self.get_product_stocks.create_integration_actions(data, object_list)
        self.assertEqual(mock_create_batch_objects.call_count, 1)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductStocks, 'get_stocks')
    def test_get_product_stock(self, mock_get_stocks, mock_get_instance):
        products = self.sample_products
        mock_get_stocks.return_value = []
        result = self.get_product_stocks.get_product_stock(products)
        self.assertEqual(result, products)
        self.assertEqual(len(self.get_product_stocks.failed_object_list), 1)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductStocks, 'get_stocks')
    def test_get_product_stock_with_failed_product(
        self,
        mock_get_stocks,
        mock_get_instance
    ):
        products = self.sample_products
        products[1].failed_reason_type = FailedReasonType.channel_app.value
        mock_get_stocks.return_value = []
        result = self.get_product_stocks.get_product_stock(products)
        self.assertEqual(result, products)
        self.assertEqual(
            len(self.get_product_stocks.failed_object_list),
            1
        )
        self.assertEqual(
            self.get_product_stocks.failed_object_list[0][0],
            products[0]
        )
        self.assertEqual(
            self.get_product_stocks.failed_object_list[0][1],
            ContentType.product.value
        )
        self.assertEqual(
            self.get_product_stocks.failed_object_list[0][2],
            "StockNotFound"
        )

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductStocks, 'get_stocks')
    def test_get_product_stock_with_stock_not_found(
        self,
        mock_get_stocks,
        mock_get_instance
    ):
        products = self.sample_products
        mock_get_stocks.return_value = []
        result = self.get_product_stocks.get_product_stock(products)
        self.assertEqual(result, products)
        self.assertEqual(
            len(self.get_product_stocks.failed_object_list),
            1
        )
        self.assertEqual(
            self.get_product_stocks.failed_object_list[0][0],
            products[0]
        )
        self.assertEqual(
            self.get_product_stocks.failed_object_list[0][1],
            ContentType.product.value
        )
        self.assertEqual(
            self.get_product_stocks.failed_object_list[0][2],
            "StockNotFound"
        )

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelProductStockEndpoint, '_list')
    def test_get_stocks(self, mock_endpoint, mock_get_instance):
        example_response = MagicMock()
        example_response.json.return_value = [
            {
                'id': 1,
                'product': 1,
                'stock': 1,
            },
            {
                'id': 2,
                'product': 2,
                'stock': 1,
            }
        ]
        mock_endpoint.return_value = example_response
        result = self.get_product_stocks.get_stocks(
            chunk=["1", "2"],
            endpoint=ChannelProductStockEndpoint()
        )
        self.assertEqual(len(result), 2)


class TestGetProductStocksWithOutCommit(TestGetProductStocks):
    pass


class TestGetProductCategoryNodes(BaseTestCaseMixin):
    """
    Test case for GetProductCategoryNodes
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetProductCategoryNodes
    """

    def setUp(self) -> None:
        self.get_product_category_nodes = GetProductCategoryNodes(
            integration=self.mock_integration
        )
        self.sample_products = [
            Product(pk=1),
            Product(pk=2)
        ]

    @patch.object(GetProductCategoryNodes, 'get_product_category')
    def test_get_data(self, mock_get_product_category):
        self.get_product_category_nodes.objects = self.sample_products
        mock_get_product_category.return_value = self.sample_products
        result = self.get_product_category_nodes.get_data()
        self.assertEqual(len(result), 2)
        self.assertEqual(result, self.sample_products)
        mock_get_product_category.assert_called_once()

    @patch.object(GetProductCategoryNodes, 'create_batch_objects')
    @patch.object(GetProductCategoryNodes, 'update_batch_request')
    def test_normalize_response(
        self,
        mock_update_batch_request,
        mock_create_batch_objects
    ):
        data = self.sample_products
        response = MagicMock()
        self.get_product_category_nodes.failed_object_list = [
            (
                self.sample_products[0],
                ContentType.product.value,
                "ProductCategoryNotFound"
            )
        ]
        self.get_product_category_nodes.normalize_response(data, response)
        mock_create_batch_objects.assert_called_once_with(
            data=[self.sample_products[0]],
            content_type=ContentType.product.value,
        )
        mock_update_batch_request.assert_called_once_with(
            mock_create_batch_objects.return_value
        )

    def test_get_product_category(self):
        products = self.sample_products
        category_tree_id = 1
        category_tree = MagicMock()
        category_tree.category_root = {"path": "/root/category"}
        category_tree_endpoint = MagicMock()
        category_tree_endpoint.retrieve.return_value = category_tree
        product_category_endpoint = MagicMock()
        product_category_endpoint.list.return_value = [
            MagicMock(
                category={
                    "path": "/root/category/category1"
                }
            ),
            MagicMock(
                category={
                    "path": "/root/category/category2"
                }
            ),
        ]
        product_category_endpoint.iterator = [
            [MagicMock(
                category={
                    "path": "/root/category/category1"
                }
            )],
            [MagicMock(
                category={
                    "path": "/root/category/category2"
                }
            )],
        ]

        with patch.object(
                ChannelCategoryTreeEndpoint,
                '__new__',
                return_value=category_tree_endpoint,
        ), patch.object(
            ChannelProductCategoryEndpoint,
            '__new__',
            return_value=product_category_endpoint,
        ):
            self.get_product_category_nodes.get_product_category(products)

        self.assertEqual(len(products), 2)
        self.assertEqual(
            self.get_product_category_nodes.failed_object_list,
            []
        )

    def test_get_product_category_with_empty_products(self):
        products = []
        result = self.get_product_category_nodes.get_product_category(products)
        self.assertEqual(result, [])
        self.assertEqual(self.get_product_category_nodes.failed_object_list, [])


class TestGetProductCategoryNodesWithIntegrationAction(TestGetProductCategoryNodes):
    """
    Test case for GetProductCategoryNodes
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetProductCategoryNodesWithIntegrationAction
    """

    def setUp(self) -> None:
        self.get_product_category_nodes = GetProductCategoryNodesWithIntegrationAction(
            integration=self.mock_integration
        )
        self.sample_products = [
            Product(
                pk=1,
                category_nodes=[
                    {
                        'pk': 1,
                        'path': '/root/category/category1'
                    }
                ]
            ),
            Product(
                pk=2,
                category_nodes=[
                    {
                        'pk': 2,
                        'path': '/root/category/category2'
                    }
                ]
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    @patch.object(
        GetProductCategoryNodesWithIntegrationAction,
        'get_data'
    )
    @patch.object(
        GetProductCategoryNodesWithIntegrationAction,
        'get_category_node_integration_action'
    )
    def test_get_data(
        self,
        mock_get_category_node_integration_action,
        mock_get_data,
        mock_get_instance
    ):
        self.get_product_category_nodes.objects = self.sample_products
        mock_get_data.return_value = self.sample_products
        result = self.get_product_category_nodes.get_data()
        self.assertEqual(len(result), 2)

    @patch.object(
        GetProductCategoryNodesWithIntegrationAction,
        'get_category_node_integration_action'
    )
    def test_get_category_node_integration_action(
        self,
        mock_get_category_node_integration_action
    ):
        mock_get_category_node_integration_action.return_value = None
        result = self.get_product_category_nodes.get_category_node_integration_action(
            self.sample_products
        )
        self.assertIsNone(result)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, 'list')
    def test_get_category_node_integration_action_with_products(
        self,
        mock_endpoint,
        mock_get_instance
    ):
        integration_action_endpoint = MagicMock()
        integration_action_endpoint.list.return_value = [
            MagicMock(
                channel=1,
                object_id=1
            ),
        ]
        integration_action_endpoint.iterator = [
            [MagicMock(
                channel=1,
                object_id=1
            )],
        ]
        mock_endpoint.return_value = integration_action_endpoint

        with patch.object(
                ChannelIntegrationActionEndpoint,
                '__new__',
                return_value=integration_action_endpoint,
                channel_id=1
        ):
            result = self.get_product_category_nodes.get_category_node_integration_action(
                self.sample_products
            )

        self.assertEqual(len(result), 2)


class TestGetBatchRequests(BaseTestCaseMixin):
    """
    Test case for GetBatchRequests
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetBatchRequests
    """

    def setUp(self):
        self.get_batch_requests = GetBatchRequests(
            integration=self.mock_integration,
            params={"limit": 1}
        )

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelBatchRequestEndpoint, 'list')
    def test_get_data(self, mock_list, mock_get_instance):
        self.get_batch_requests.params = {
            "limit": 1
        }
        mock_list.return_value = [
            {
                "pk": 22,
                "channel": 6,
                "local_batch_id": "89bb31b0-6700-4a9d-8bae-308ac938649c",
                "remote_batch_id": "028b8e66-2a4c-45b5-912f-9ab90036c78a",
                "content_type": "product",
                "status": "sent_to_remote"
            }
        ]

        data = self.get_batch_requests.get_data()

        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].get("pk"), 22)
        self.assertEqual(data[0].get("channel"), 6)
        self.assertEqual(data[0].get("local_batch_id"), "89bb31b0-6700-4a9d-8bae-308ac938649c")
        self.assertEqual(data[0].get("remote_batch_id"), "028b8e66-2a4c-45b5-912f-9ab90036c78a")
        self.assertEqual(data[0].get("content_type"), "product")
        self.assertEqual(data[0].get("status"), "sent_to_remote")
        mock_list.assert_called_once()


class TestGetUpdatedProductStocks(BaseTestCaseMixin):
    """
    Test case for GetUpdatedProductStocks
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetUpdatedProductStocks
    """
    def setUp(self) -> None:
        self.get_updated_product_stocks = GetUpdatedProductStocks(
            integration=self.mock_integration
        )
        self.sample_stocks = [
            ProductStock(
                pk=3,
                product=1,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            ),
            ProductStock(
                pk=4,
                product=2,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetUpdatedProductStocks, 'get_product_stocks')
    @patch.object(GetUpdatedProductStocks, 'get_stocks_with_available')
    def test_get_data(
        self,
        mock_get_stocks_with_available,
        mock_get_product_stocks,
        mock_get_instance
    ):
        mock_get_product_stocks.return_value = self.sample_stocks
        mock_get_stocks_with_available.return_value = self.sample_stocks

        stocks = self.get_updated_product_stocks.get_data()

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)
        mock_get_product_stocks.assert_called_once()
        mock_get_stocks_with_available.assert_called_once()

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelProductStockEndpoint, 'list')
    @patch.object(OmnitronCommandInterface, 'create_batch_objects')
    @patch.object(OmnitronCommandInterface, 'update_batch_request')
    def test_get_product_stocks(
        self,
        mock_update_batch_request,
        mock_create_batch_objects,
        mock_endpoint,
        mock_get_instance
    ):
        mock_endpoint.return_value = [
            {
                'pk': 3,
                'product': 1,
                'stock': 0,
                'stock_list': 4,
                'unit_type': 'qty',
                'extra_field': {},
                'sold_quantity_unreported': 0,
                'modified_date': '2023-12-19T08:38:48.476005Z',
                'created_date': '2023-12-19T08:38:48.475992Z'
            },
            {
                'pk': 4,
                'product': 2,
                'stock': 0,
                'stock_list': 4,
                'unit_type': 'qty',
                'extra_field': {},
                'sold_quantity_unreported': 0,
                'modified_date': '2023-12-19T08:38:48.476005Z',
                'created_date': '2023-12-19T08:38:48.475992Z'
            }
        ]
        mock_create_batch_objects.return_value = [
            {
                'pk': 3,
                'failed_reason_type': None,
                'remote_id': 3,
                'version_date': '2023-12-19T08:38:48.476005Z',
                'content_type': 'product_stock',
            },
            {
                'pk': 4,
                'failed_reason_type': None,
                'remote_id': 4,
                'version_date': '2023-12-19T08:38:48.476005Z',
                'content_type': 'product_stock',
            }
        ]

        stocks = self.get_updated_product_stocks.get_product_stocks()
        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].get('pk'), 3)
        self.assertEqual(stocks[1].get('pk'), 4)
        mock_endpoint.assert_called_once()
        mock_create_batch_objects.assert_called_once()
        mock_update_batch_request.assert_called_once()

    @patch.object(BaseClient, 'get_instance')
    def test_get_stocks_with_available(
        self,
        mock_get_instance
    ):
        mock_endpoint = MagicMock(channel_id=1)
        mock_endpoint.list.return_value = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_stock.value,
                object_id=self.sample_stocks[0].pk,
                remote_id=None,
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_stock.value,
                object_id=self.sample_stocks[1].pk,
                remote_id=None,
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

        mock_endpoint.iterator = [
            [
                MagicMock(
                    pk=1,
                    channel=2,
                    content_type=ContentType.product_stock.value,
                    object_id=self.sample_stocks[0].pk,
                    remote_id=None,
                    version_date="2023-12-28T10:28:17.186730Z",
                    state={},
                    modified_date="2023-12-28T10:28:17.187032Z",
                    local_batch_id=None,
                    status=None,
                    created_date="2023-12-28T10:28:17.187014Z"
                )
            ],
            [
                MagicMock(
                    pk=2,
                    channel=2,
                    content_type=ContentType.product_stock.value,
                    object_id=self.sample_stocks[1].pk,
                    remote_id=None,
                    version_date="2023-12-28T10:28:17.186730Z",
                    state={},
                    modified_date="2023-12-28T10:28:17.187032Z",
                    local_batch_id=None,
                    status=None,
                    created_date="2023-12-28T10:28:17.187014Z"
                )
            ]
        ]

        with patch.object(
                ChannelIntegrationActionEndpoint,
                '__new__',
                return_value=mock_endpoint,
        ):
            stocks = self.get_updated_product_stocks.get_stocks_with_available(
                self.sample_stocks
            )

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)


class TestGetInsertedProductStocks(BaseTestCaseMixin):
    """
    Test case for GetInsertedProductStocks
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedProductStocks
    """

    def setUp(self):
        self.get_inserted_product_stocks = GetInsertedProductStocks(
            integration=self.mock_integration
        )
        self.sample_stocks = [
            ProductStock(
                pk=3,
                product=1,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            ),
            ProductStock(
                pk=4,
                product=2,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, 'list')
    def test_get_stocks_with_available(
        self,
        mock_endpoint,
        mock_get_instance
    ):
        example_response = MagicMock()
        example_response.return_value = [
            {
                "pk": 1,
                "channel": 2,
                "content_type": {
                    "id": 119,
                    "app_label": "search",
                    "model": "productstock"
                },
                "object_id": self.sample_stocks[0].pk,
                "remote_id": None,
                "version_date": "2023-12-28T10:28:17.186730Z",
                "state": {},
                "modified_date": "2023-12-28T10:28:17.187032Z",
                "local_batch_id": None,
                "status": None,
                "created_date": "2023-12-28T10:28:17.187014Z"
            },
            {
                "pk": 2,
                "channel": 2,
                "content_type": {
                    "id": 119,
                    "app_label": "search",
                    "model": "productstock"
                },
                "object_id": self.sample_stocks[1].pk,
                "remote_id": None,
                "version_date": "2023-12-28T10:28:17.186730Z",
                "state": {},
                "modified_date": "2023-12-28T10:28:17.187032Z",
                "local_batch_id": None,
                "status": None,
                "created_date": "2023-12-28T10:28:17.187014Z"
            }
        ]

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response,
        ):
            stocks = self.get_inserted_product_stocks.get_stocks_with_available(
                self.sample_stocks
            )

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)


class TestGetUpdatedProductStocksFromExtraStockList(BaseTestCaseMixin):
    """
    Test case for GetupdatedProductStocksFromExtraStockList
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetUpdatedProductStocksFromExtraStockList
    """

    def setUp(self) -> None:
        self.get_updated_product_stocks_from_extra_stock_list = GetUpdatedProductStocksFromExtraStockList(
            integration=self.mock_integration
        )
        self.sample_stocks = [
            ProductStock(
                pk=3,
                product=1,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            ),
            ProductStock(
                pk=4,
                product=2,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            )
        ]
        self.objects = self.sample_stocks
        self.stock_list_id = 4

    @patch.object(
        BaseClient, 
        'get_instance'
    )
    @patch.object(
        GetUpdatedProductStocksFromExtraStockList, 
        'get_product_stocks'
    )
    @patch.object(
        GetUpdatedProductStocksFromExtraStockList, 
        'get_integration_actions'
    )
    def test_get_data(
        self,
        mock_get_integration_actions,
        mock_get_product_stocks,
        mock_get_instance
    ):
        mock_get_integration_actions.return_value = self.sample_stocks
        mock_get_product_stocks.return_value = self.sample_stocks
        stocks = self.get_updated_product_stocks_from_extra_stock_list.get_data()

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)

    @patch.object(
        BaseClient, 
        'get_instance'
    )
    @patch.object(
        ChannelIntegrationActionEndpoint, 
        'list'
    )
    @patch.object(
        GetUpdatedProductStocksFromExtraStockList, 
        'create_batch_objects'
    )
    @patch.object(
        GetUpdatedProductStocksFromExtraStockList, 
        'update_batch_request'
    )
    def test_get_product_stocks(
        self,
        mock_update_batch_request,
        mock_create_batch_objects,
        mock_list,
        mock_get_instance
    ):
        self.get_updated_product_stocks_from_extra_stock_list.objects = self.objects
        self.get_updated_product_stocks_from_extra_stock_list.stock_list_id = self.stock_list_id

        mock_endpoint = MagicMock()
        mock_endpoint.list.return_value = [
            {
                'pk': 3,
                'product': 1,
                'stock': 0,
                'stock_list': 4,
                'unit_type': 'qty',
                'extra_field': {},
                'sold_quantity_unreported': 0,
                'modified_date': '2023-12-19T08:38:48.476005Z',
                'created_date': '2023-12-19T08:38:48.475992Z'
            },
            {
                'pk': 4,
                'product': 2,
                'stock': 0,
                'stock_list': 4,
                'unit_type': 'qty',
                'extra_field': {},
                'sold_quantity_unreported': 0,
                'modified_date': '2023-12-19T08:38:48.476005Z',
                'created_date': '2023-12-19T08:38:48.475992Z'
            }
        ]

        with patch.object(
            ChannelExtraProductStockEndpoint,
            '__new__',
            return_value=mock_endpoint
        ):
            stocks = self.get_updated_product_stocks_from_extra_stock_list.get_product_stocks()

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].get('pk'), 3)
        self.assertEqual(stocks[1].get('pk'), 4)

    def test_get_integration_actions_without_stocks(self):
        self.sample_stocks = []
        stocks = self.get_updated_product_stocks_from_extra_stock_list.get_integration_actions(
            self.sample_stocks
        )
        self.assertEqual(stocks, [])

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, 'list')
    def test_get_integration_actions(
        self,
        mock_get_instance,
        mock_list
    ):
        mock_endpoint = MagicMock()
        mock_endpoint.list.return_value = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_stock.value,
                object_id=self.sample_stocks[0].pk,
                remote_id=None,
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_stock.value,
                object_id=self.sample_stocks[1].pk,
                remote_id=None,
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]
        mock_endpoint.iterator = [
            [
                MagicMock(
                    pk=1,
                    channel=2,
                    content_type=ContentType.product_stock.value,
                    object_id=self.sample_stocks[0].pk,
                    remote_id=None,
                    version_date="2023-12-28T10:28:17.186730Z",
                    state={},
                    modified_date="2023-12-28T10:28:17.187032Z",
                    local_batch_id=None,
                    status=None,
                    created_date="2023-12-28T10:28:17.187014Z"
                )
            ],
            [
                MagicMock(
                    pk=2,
                    channel=2,
                    content_type=ContentType.product_stock.value,
                    object_id=self.sample_stocks[1].pk,
                    remote_id=None,
                    version_date="2023-12-28T10:28:17.186730Z",
                    state={},
                    modified_date="2023-12-28T10:28:17.187032Z",
                    local_batch_id=None,
                    status=None,
                    created_date="2023-12-28T10:28:17.187014Z"
                )
            ]
        ]

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=mock_endpoint
        ):
            stocks = self.get_updated_product_stocks_from_extra_stock_list.get_integration_actions(
                self.sample_stocks
            )

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)


class TestGetProductPricesFromProductStocks(BaseTestCaseMixin):
    """
    Test case for GetProductPricesFromProductStocks
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetProductPricesFromProductStocks
    """

    def setUp(self) -> None:
        self.get_product_prices_from_product_stocks = GetProductPricesFromProductStocks(
            integration=self.mock_integration
        )
        self.sample_stocks = [
            ProductStock(
                pk=3,
                product=1,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z',
                productprice=10
            ),
            ProductStock(
                pk=4,
                product=2,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z',
                productprice=10
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductPricesFromProductStocks, 'get_product_price')
    def test_get_data(
        self,
        mock_get_product_price,
        mock_get_instance
    ):
        self.get_product_prices_from_product_stocks.objects = self.sample_stocks
        stocks = self.get_product_prices_from_product_stocks.get_data()
        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductPricesFromProductStocks, 'create_batch_objects')
    @patch.object(GetProductPricesFromProductStocks, 'create_integration_actions')
    @patch.object(GetProductPricesFromProductStocks, 'update_batch_request')
    def test_normalize_response(
        self,
        mock_update_batch_request,
        mock_create_integration_actions,
        mock_create_batch_objects,
        mock_get_instance
    ):
        data = self.get_product_prices_from_product_stocks.normalize_response(
            self.sample_stocks,
            None
        )
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(data[1].pk, self.sample_stocks[1].pk)

    @patch.object(GetProductPricesFromProductStocks, 'create_batch_objects')
    def test_create_integration_actions(
        self,
        mock_create_batch_objects
    ):
        self.get_product_prices_from_product_stocks.create_integration_actions(
            self.sample_stocks,
            []
        )
        self.assertEqual(mock_create_batch_objects.call_count, 1)

    def test_get_product_price_with_empty_stock(self):
        stocks = self.get_product_prices_from_product_stocks.get_product_price(
            []
        )
        self.assertEqual(stocks, [])

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelExtraProductPriceEndpoint, 'list')
    @patch.object(GetProductPricesFromProductStocks, 'get_prices')
    def test_get_product_price(
        self,
        mock_get_prices,
        mock_list,
        mock_get_instance
    ):
        example_response = MagicMock()
        example_response.list.return_value = self.sample_stocks
        example_response.iterator = iter(self.sample_stocks)

        with patch.object(
            ChannelExtraProductPriceEndpoint,
            '__new__',
            return_value=example_response
        ):
            stocks = self.get_product_prices_from_product_stocks.get_product_price(
                self.sample_stocks
            )

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.sample_stocks[0].pk)
        self.assertEqual(stocks[1].pk, self.sample_stocks[1].pk)
        self.assertEqual(
            stocks[0].failed_reason_type, 
            FailedReasonType.channel_app.value
        )
        self.assertEqual(
            stocks[1].failed_reason_type, 
            FailedReasonType.channel_app.value
        )

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelExtraProductPriceEndpoint, 'list')
    def test_get_prices(
        self,
        mock_list,
        mock_get_instance
    ):
        product_prices = [
            MagicMock(
                pk=1,
                product=5,
                price='645670668.11',
                price_list=1,
                currency_type='try',
                tax_rate='32.61',
                retail_price=None,
                extra_field={},
                discount_percentage='63.57',
                modified_date='2024-01-02T07:53:06.392780Z',
                created_date='2024-01-02T07:53:06.392771Z',
                price_type='default'
            )
        ]
        
        example_response = MagicMock()
        example_response.list.return_value = product_prices
        example_response.iterator = iter(product_prices)

        with patch.object(
            ChannelExtraProductPriceEndpoint,
            '__new__',
            return_value=example_response
        ):
            stocks = self.get_product_prices_from_product_stocks.get_prices(
                "1,2,3",
                ChannelExtraProductPriceEndpoint()
            )
        
        self.assertEqual(len(stocks), 1)
        self.assertEqual(stocks[0].pk, product_prices[0].pk)
        self.assertEqual(stocks[0].product, product_prices[0].product)
        self.assertEqual(stocks[0].price, product_prices[0].price)


class TestGetProductStocksFromProductPrices(BaseTestCaseMixin):
    """
    Test case for GetProductStocksFromProductPrices
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetProductStocksFromProductPrices
    """

    def setUp(self) -> None:
        self.stocks = [
            MagicMock(
                pk=3,
                product=1,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            ),
        ]
        self.prices = [
            MagicMock(
                pk=1,
                product=5,
                price='645670668.11',
                price_list=1,
                currency_type='try',
                tax_rate='32.61',
                retail_price=None,
                extra_field={},
                discount_percentage='63.57',
                modified_date='2024-01-02T07:53:06.392780Z',
                created_date='2024-01-02T07:53:06.392771Z',
                price_type='default',
                failed_reason_type=FailedReasonType.channel_app.value
            )
        ]
        self.get_product_stocks_from_product_prices = GetProductStocksFromProductPrices(
            integration=self.mock_integration
        )

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetProductStocksFromProductPrices, 'get_product_stock')
    def test_get_data(self, mock_get_instance, mock_get_product_stock):
        self.get_product_stocks_from_product_prices.objects = self.prices
        self.assertEqual(
            self.get_product_stocks_from_product_prices.get_data(),
            self.prices
        )

    @patch.object(
        BaseClient, 
        'get_instance'
    )
    @patch.object(
        GetProductStocksFromProductPrices, 
        'create_batch_objects'
    )
    @patch.object(
        GetProductStocksFromProductPrices, 
        'create_integration_actions'
    )
    @patch.object(
        GetProductStocksFromProductPrices, 
        'update_batch_request'
    )
    def test_normalize_response(
        self,
        mock_update_batch_request,
        mock_create_integration_actions,
        mock_create_batch_objects,
        mock_get_instance
    ):
        data = self.get_product_stocks_from_product_prices.normalize_response(
            self.stocks,
            None
        )
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].pk, self.stocks[0].pk)

    @patch.object(GetProductStocksFromProductPrices, 'create_batch_objects')
    def test_create_integration_actions(self, mock_create_batch_objects):
        self.get_product_stocks_from_product_prices.create_integration_actions(
            self.stocks,
            []
        )
        self.assertEqual(mock_create_batch_objects.call_count, 1)

    def test_get_product_stock_with_empty_stock(self):
        self.stocks = []
        stocks = self.get_product_stocks_from_product_prices.get_product_stock(
            self.stocks
        )
        self.assertEqual(stocks, [])

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelExtraProductPriceEndpoint, 'list')
    @patch.object(GetProductStocksFromProductPrices, 'get_stocks')
    def test_get_product_stock(self, mock_get_stocks, mock_list, mock_get_instance):
        mock_get_stocks.return_value = self.stocks
        prices = self.get_product_stocks_from_product_prices.get_product_stock(
            self.prices
        )
        self.assertEqual(len(prices), 1)
        self.assertEqual(prices[0].pk, self.prices[0].pk)
        self.assertEqual(prices[0].product, self.prices[0].product)
        self.assertEqual(prices[0].price, self.prices[0].price)
        self.assertEqual(prices[0].failed_reason_type, FailedReasonType.channel_app.value)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelExtraProductPriceEndpoint, 'list')
    def test_get_stocks(self, mock_list, mock_get_instance):
        example_response = MagicMock()
        example_response.list.return_value = self.stocks
        example_response.iterator = iter(self.stocks)

        with patch.object(
            ChannelExtraProductStockEndpoint,
            '__new__',
            return_value=example_response
        ):
            stocks = self.get_product_stocks_from_product_prices.get_stocks(
                "1,2,3",
                ChannelExtraProductStockEndpoint()
            )

        self.assertEqual(len(stocks), 1)
        self.assertEqual(stocks[0].pk, self.stocks[0].pk)
        self.assertEqual(stocks[0].product, self.stocks[0].product)
        self.assertEqual(stocks[0].stock, self.stocks[0].stock)


class TestGetInsertedProductStocksFromExtraStockList(BaseTestCaseMixin):
    """
    Test case for GetInsertedProductStocksFromExtraStockList
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedProductStocksFromExtraStockList
    """

    def setUp(self) -> None:
        self.get_inserted_product_stocks_from_extra_stock_list = GetInsertedProductStocksFromExtraStockList(
            integration=self.mock_integration
        )
        self.sample_stocks = [
            ProductStock(
                pk=3,
                product=1,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                remote_id=3,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            ),
            ProductStock(
                pk=4,
                product=2,
                stock=0,
                stock_list=4,
                unit_type='qty',
                extra_field={},
                sold_quantity_unreported=0,
                remote_id=4,
                modified_date='2023-12-19T08:38:48.476005Z',
                created_date='2023-12-19T08:38:48.475992Z'
            )
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_stock.value,
                object_id=self.sample_stocks[0].pk,
                remote_id=None,
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_stock.value,
                object_id=self.sample_stocks[1].pk,
                remote_id=None,
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, 'list')
    def test_get_integration_actions(self, mock_list, mock_get_instance):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            stocks = self.get_inserted_product_stocks_from_extra_stock_list.get_integration_actions(
                self.sample_stocks
            )

        self.assertEqual(len(stocks), 2)
        self.assertEqual(stocks[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(stocks[1].pk, self.integration_actions[1].object_id)
        self.assertEqual(stocks[0].remote_id, self.sample_stocks[0].remote_id)
        self.assertEqual(stocks[1].remote_id, self.sample_stocks[1].remote_id)


class TestGetUpdatedProductPrices(BaseTestCaseMixin):
    """
    Test case for GetUpdatedProductPrices
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetUpdatedProductPrices
    """

    def setUp(self) -> None:
        self.get_updated_product_prices = GetUpdatedProductPrices(
            integration=self.mock_integration
        )
        self.product_prices = [
            ProductPrice(
                pk=1,
                product=5,
                price='645670668.11',
                price_list=1,
                currency_type='try',
                tax_rate='32.61',
                retail_price=None,
                extra_field={},
                discount_percentage='63.57',
                modified_date='2024-01-02T07:53:06.392780Z',
                created_date='2024-01-02T07:53:06.392771Z',
                price_type='default'
            ),
            ProductPrice(
                pk=2,
                product=6,
                price='493154528.49',
                price_list=2,
                currency_type='try',
                tax_rate='31.13',
                retail_price=None,
                extra_field={},
                discount_percentage='26.55',
                modified_date='2024-01-09T13:20:23.472505Z',
                created_date='2024-01-09T13:20:23.472485Z',
                price_type='default'
            )
        ]
        self.product_prices_json = [
            {
                "pk": 1,
                "product": 5,
                "price": "645670668.11",
                "price_list": 1,
                "currency_type": "try",
                "tax_rate": "32.61",
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "63.57",
                "modified_date": "2024-01-02T07:53:06.392780Z",
                "created_date": "2024-01-02T07:53:06.392771Z",
                "price_type": "default"
            },
            {
                "pk": 2,
                "product": 6,
                "price": "493154528.49",
                "price_list": 2,
                "currency_type": "try",
                "tax_rate": "31.13",
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "26.55",
                "modified_date": "2024-01-09T13:20:23.472505Z",
                "created_date": "2024-01-09T13:20:23.472485Z",
                "price_type": "default"
            }
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[0].pk,
                remote_id='1',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[1].pk,
                remote_id='2',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    @patch.object(
        GetUpdatedProductPrices, 
        'get_product_prices'
    )
    @patch.object(
        GetUpdatedProductPrices, 
        'get_integration_actions'
    )
    def test_get_data(
        self, 
        mock_get_product_prices, 
        mock_get_integration_actions
    ):
        mock_get_product_prices.return_value = self.product_prices
        mock_get_integration_actions.return_value = self.integration_actions
        data = self.get_updated_product_prices.get_data()
        
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.product_prices[0].pk)
        self.assertEqual(data[1].pk, self.product_prices[1].pk)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetUpdatedProductPrices, 'create_batch_objects')
    @patch.object(GetUpdatedProductPrices, 'update_batch_request')
    def test_get_product_prices(
        self, 
        mock_get_instance, 
        mock_create_batch_objects, 
        mock_update_batch_request
    ):
        example_response = MagicMock()
        example_response.list.return_value = self.product_prices_json

        with patch.object(
            ChannelProductPriceEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_updated_product_prices.get_product_prices()

        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].get('pk'), self.product_prices[0].pk)
        self.assertEqual(data[1].get('pk'), self.product_prices[1].pk)

    @patch.object(BaseClient, 'get_instance')
    def test_get_integration_actions(self, mock_get_instance):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_updated_product_prices.get_integration_actions(
                self.product_prices
            )
        
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(data[1].pk, self.integration_actions[1].object_id)
        self.assertEqual(
            data[0].remote_id, 
            self.integration_actions[0].remote_id
        )
        self.assertEqual(
            data[1].remote_id, 
            self.integration_actions[1].remote_id
            )
        

class TestGetInsertedProductPrices(BaseTestCaseMixin):
    """
    Test case for GetInsertedProductPrices
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedProductPrices
    """
    def setUp(self) -> None:
        self.get_inserted_product_prices = GetInsertedProductPrices(
            integration=self.mock_integration
        )
        self.product_prices = [
            ProductPrice(
                pk=1,
                product=5,
                price='645670668.11',
                price_list=1,
                currency_type='try',
                tax_rate='32.61',
                remote_id='1',
                retail_price=None,
                extra_field={},
                discount_percentage='63.57',
                modified_date='2024-01-02T07:53:06.392780Z',
                created_date='2024-01-02T07:53:06.392771Z',
                price_type='default'
            ),
            ProductPrice(
                pk=2,
                product=6,
                price='493154528.49',
                price_list=2,
                currency_type='try',
                tax_rate='31.13',
                remote_id='2',
                retail_price=None,
                extra_field={},
                discount_percentage='26.55',
                modified_date='2024-01-09T13:20:23.472505Z',
                created_date='2024-01-09T13:20:23.472485Z',
                price_type='default'
            )
        ]
        self.product_prices_json = [
            {
                "pk": 1,
                "product": 5,
                "price": "645670668.11",
                "price_list": 1,
                "currency_type": "try",
                "tax_rate": "32.61",
                "remote_id": None,
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "63.57",
                "modified_date": "2024-01-02T07:53:06.392780Z",
                "created_date": "2024-01-02T07:53:06.392771Z",
                "price_type": "default"
            },
            {
                "pk": 2,
                "product": 6,
                "price": "493154528.49",
                "price_list": 2,
                "currency_type": "try",
                "tax_rate": "31.13",
                "remote_id": None,
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "26.55",
                "modified_date": "2024-01-09T13:20:23.472505Z",
                "created_date": "2024-01-09T13:20:23.472485Z",
                "price_type": "default"
            }
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[0].pk,
                remote_id='1',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[1].pk,
                remote_id='2',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    def test_get_integration_actions(self, mock_get_instance):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_inserted_product_prices.get_integration_actions(
                self.product_prices
            )
        
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(data[1].pk, self.integration_actions[1].object_id)
        self.assertEqual(
            data[0].remote_id, 
            self.integration_actions[0].remote_id
        )
        self.assertEqual(
            data[1].remote_id, 
            self.integration_actions[1].remote_id
            )
        

class TestGetInsertedProductPricesFromExtraPriceList(BaseTestCaseMixin):
    """
    Test case for GetInsertedProductPricesFromExtraPriceList
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedProductPricesFromExtraPriceList
    """
    def setUp(self) -> None:
        self.get_inserted_product_prices_from_extra_price_list = GetInsertedProductPricesFromExtraPriceList(
            integration=self.mock_integration
        )
        self.product_prices = [
            ProductPrice(
                pk=1,
                product=5,
                price='645670668.11',
                price_list=1,
                currency_type='try',
                tax_rate='32.61',
                remote_id='1',
                retail_price=None,
                extra_field={},
                discount_percentage='63.57',
                modified_date='2024-01-02T07:53:06.392780Z',
                created_date='2024-01-02T07:53:06.392771Z',
                price_type='default'
            ),
            ProductPrice(
                pk=2,
                product=6,
                price='493154528.49',
                price_list=2,
                currency_type='try',
                tax_rate='31.13',
                remote_id='2',
                retail_price=None,
                extra_field={},
                discount_percentage='26.55',
                modified_date='2024-01-09T13:20:23.472505Z',
                created_date='2024-01-09T13:20:23.472485Z',
                price_type='default'
            )
        ]
        self.product_prices_json = [
            {
                "pk": 1,
                "product": 5,
                "price": "645670668.11",
                "price_list": 1,
                "currency_type": "try",
                "tax_rate": "32.61",
                "remote_id": None,
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "63.57",
                "modified_date": "2024-01-02T07:53:06.392780Z",
                "created_date": "2024-01-02T07:53:06.392771Z",
                "price_type": "default"
            },
            {
                "pk": 2,
                "product": 6,
                "price": "493154528.49",
                "price_list": 2,
                "currency_type": "try",
                "tax_rate": "31.13",
                "remote_id": None,
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "26.55",
                "modified_date": "2024-01-09T13:20:23.472505Z",
                "created_date": "2024-01-09T13:20:23.472485Z",
                "price_type": "default"
            }
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[0].pk,
                remote_id='1',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[1].pk,
                remote_id='2',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    def test_get_integration_actions(self):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_inserted_product_prices_from_extra_price_list.get_integration_actions(
                self.product_prices
            )

        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(data[1].pk, self.integration_actions[1].object_id)
        self.assertEqual(
            data[0].remote_id, 
            self.integration_actions[0].remote_id
        )
        self.assertEqual(
            data[1].remote_id, 
            self.integration_actions[1].remote_id
            )
        

class TestGetUpdatedProductPricesFromExtraPriceList(BaseTestCaseMixin):
    """
    Test case for GetUpdatedProductPricesFromExtraPriceList
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetUpdatedProductPricesFromExtraPriceList
    """
    def setUp(self) -> None:
        self.get_updated_product_prices_from_extra_price_list = GetUpdatedProductPricesFromExtraPriceList(
            integration=self.mock_integration
        )
        self.get_updated_product_prices_from_extra_price_list.objects = [1, 2]
        self.get_updated_product_prices_from_extra_price_list.price_list_id = 1
        self.product_prices = [
            ProductPrice(
                pk=1,
                product=5,
                price='645670668.11',
                price_list=1,
                currency_type='try',
                tax_rate='32.61',
                remote_id='1',
                retail_price=None,
                extra_field={},
                discount_percentage='63.57',
                modified_date='2024-01-02T07:53:06.392780Z',
                created_date='2024-01-02T07:53:06.392771Z',
                price_type='default'
            ),
            ProductPrice(
                pk=2,
                product=6,
                price='493154528.49',
                price_list=1,
                currency_type='try',
                tax_rate='31.13',
                remote_id='2',
                retail_price=None,
                extra_field={},
                discount_percentage='26.55',
                modified_date='2024-01-09T13:20:23.472505Z',
                created_date='2024-01-09T13:20:23.472485Z',
                price_type='default'
            )
        ]
        self.product_prices_json = [
            {
                "pk": 1,
                "product": 5,
                "price": "645670668.11",
                "price_list": 1,
                "currency_type": "try",
                "tax_rate": "32.61",
                "remote_id": None,
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "63.57",
                "modified_date": "2024-01-02T07:53:06.392780Z",
                "created_date": "2024-01-02T07:53:06.392771Z",
                "price_type": "default"
            },
            {
                "pk": 2,
                "product": 6,
                "price": "493154528.49",
                "price_list": 1,
                "currency_type": "try",
                "tax_rate": "31.13",
                "remote_id": None,
                "retail_price": None,
                "extra_field": {},
                "discount_percentage": "26.55",
                "modified_date": "2024-01-09T13:20:23.472505Z",
                "created_date": "2024-01-09T13:20:23.472485Z",
                "price_type": "default"
            }
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[0].pk,
                remote_id='1',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            ),
            MagicMock(
                pk=2,
                channel=2,
                content_type=ContentType.product_price.value,
                object_id=self.product_prices[1].pk,
                remote_id='2',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    @patch.object(
        GetUpdatedProductPricesFromExtraPriceList, 
        'get_product_prices'
    )
    @patch.object(
        GetUpdatedProductPricesFromExtraPriceList, 
        'get_integration_actions'
    )
    def test_get_data(
        self, 
        mock_get_product_prices, 
        mock_get_integration_actions
    ):
        mock_get_product_prices.return_value = self.product_prices
        mock_get_integration_actions.return_value = self.integration_actions
        data = self.get_updated_product_prices_from_extra_price_list.get_data()
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.product_prices[0].pk)
        self.assertEqual(data[1].pk, self.product_prices[1].pk)

    @patch.object(
        BaseClient, 
        'get_instance'
    )
    @patch.object(
        GetUpdatedProductPricesFromExtraPriceList, 
        'create_batch_objects'
    )
    @patch.object(
        GetUpdatedProductPricesFromExtraPriceList, 
        'update_batch_request'
    )
    def test_get_product_prices(
        self, 
        mock_get_instance,
        mock_create_batch_objects,
        mock_update_batch_request
    ):
        example_response = MagicMock()
        example_response.list.return_value = self.product_prices_json

        with patch.object(
            ChannelExtraProductPriceEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_updated_product_prices_from_extra_price_list.get_product_prices()

        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].get('pk'), self.product_prices[0].pk)
        self.assertEqual(data[1].get('pk'), self.product_prices[1].pk)

    def test_get_integration_actions_with_empty_prices(self):
        self.product_prices = []
        data = self.get_updated_product_prices_from_extra_price_list.get_integration_actions(
            self.product_prices
        )
        self.assertEqual(data, [])

    @patch.object(BaseClient, 'get_instance')
    def test_get_integration_actions(self, mock_get_instance):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_updated_product_prices_from_extra_price_list.get_integration_actions(
                self.product_prices
            )

        self.assertEqual(len(data), 2)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(data[1].pk, self.integration_actions[1].object_id)
        self.assertEqual(
            data[0].remote_id, 
            self.integration_actions[0].remote_id
        )
        self.assertEqual(
            data[1].remote_id, 
            self.integration_actions[1].remote_id
        )


class TestGetUpdatedProductImages(BaseTestCaseMixin):
    """
    Test case for GetUpdatedProductImages
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetUpdatedProductImages
    """

    def setUp(self) -> None:
        self.get_updated_product_images = GetUpdatedProductImages(
            integration=self.mock_integration
        )
        self.product_images = [
            ProductImage(
                pk=1,
                product=7,
                image='/media/products/2024/01/11/7/0aaa1bf0-8b4f-4b2c-b166-9d61d16879ce.jpg',
                order=0,
                source=None,
                modified_date='2024-01-11T07:31:45.520043Z',
                created_date='2024-01-11T07:31:45.520013Z',
                height=56,
                width=130,
                hash=None,
                status='active',
                is_active=True
            ),
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_image.value,
                object_id=self.product_images[0].pk,
                remote_id='1',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    @patch.object(GetUpdatedProductImages, 'get_product_images')
    @patch.object(GetUpdatedProductImages, 'get_integration_actions')
    def test_get_data(
        self, 
        mock_get_product_images, 
        mock_get_integration_actions
    ):
        mock_get_product_images.return_value = self.product_images
        mock_get_integration_actions.return_value = self.product_images
        data = self.get_updated_product_images.get_data()
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].pk, self.product_images[0].pk)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(GetUpdatedProductImages, 'create_batch_objects')
    @patch.object(GetUpdatedProductImages, 'update_batch_request')
    def test_get_product_images(
        self, 
        mock_update_batch_request, 
        mock_create_batch_objects, 
        mock_get_instance
    ):
        example_response = MagicMock()
        example_response.list.return_value = self.product_images

        with patch.object(
            ChannelProductImageEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_updated_product_images.get_product_images()

        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].pk, self.product_images[0].pk)

    @patch.object(BaseClient, 'get_instance')
    def test_get_integration_actions(
        self,
        mock_get_instance
    ):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_updated_product_images.get_integration_actions(
                self.product_images
            )

        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(
            data[0].remote_id, 
            self.integration_actions[0].remote_id
        )


class TestGetInsertedProductImages(BaseTestCaseMixin):
    """
    Test case for GetInsertedProductImages
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestGetInsertedProductImages
    """

    def setUp(self) -> None:
        self.get_inserted_product_images = GetInsertedProductImages(
            integration=self.mock_integration
        )
        self.product_images = [
            ProductImage(
                pk=1,
                product=1,
                image='/media/products/2024/01/11/7/0aaa1bf0-8b4f-4b2c-b166-9d61d16879ce.jpg',
                order=0,
                source=None,
                modified_date='2024-01-11T07:31:45.520043Z',
                created_date='2024-01-11T07:31:45.520013Z',
                height=56,
                width=130,
                hash=None,
                status='active',
                is_active=True,
                remote_id=None
            ),
        ]
        self.integration_actions = [
            MagicMock(
                pk=1,
                channel=2,
                content_type=ContentType.product_image.value,
                object_id=self.product_images[0].pk,
                remote_id='1',
                version_date="2023-12-28T10:28:17.186730Z",
                state={},
                modified_date="2023-12-28T10:28:17.187032Z",
                local_batch_id=None,
                status=None,
                created_date="2023-12-28T10:28:17.187014Z"
            )
        ]

    @patch.object(BaseClient, 'get_instance')
    def test_get_integration_actions(self, mock_get_instance):
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_inserted_product_images.get_integration_actions(
                self.product_images
            )

        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(
            data[0].remote_id, 
            self.integration_actions[0].remote_id
        )

    @patch.object(BaseClient, 'get_instance')
    def test_get_integration_actions_with_not_in_product_case(
        self, 
        mock_get_instance
    ):
        self.product_images[0].product = 0
        example_response = MagicMock()
        example_response.list.return_value = self.integration_actions
        example_response.iterator = iter(self.integration_actions)

        with patch.object(
            ChannelIntegrationActionEndpoint,
            '__new__',
            return_value=example_response
        ):
            data = self.get_inserted_product_images.get_integration_actions(
                self.product_images
            )

        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].pk, self.integration_actions[0].object_id)
        self.assertEqual(
            data[0].failed_reason_type, 
            FailedReasonType.channel_app.value
        )


class TestProcessProductBatchRequests(BaseTestCaseMixin):
    """
    Test case for ProcessProductBatchRequests
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestProcessProductBatchRequests
    """

    def setUp(self) -> None:
        self.process_product_batch_requests = ProcessProductBatchRequests(
            integration=self.mock_integration
        )
        self.batch_requests = [
            MagicMock(
                pk=1,
                channel=2,
                local_batch_id='10147300-fe33-4c1f-a7c4-d2eb4001df70',
                remote_batch_id=None,
                status='initialized',
                objects=None,
                created_date='2024-01-11T08:12:11.481482Z',
                modified_date='2024-01-11T08:12:11.481504Z'
            ),
        ]
        self.sample_data = [
            ProductBatchRequestResponseDto(
                status='initialized',
                sku='1',
                remote_id=None,
                message=None,
            )
        ]
        self.not_correct_sample_data = [
            {
                "status": "initialized",
                "sku": "1",
                "remote_id": None,
                "message": None,
            }
        ]

    def test_validated_data(self):
        data = self.process_product_batch_requests.validated_data(
            self.sample_data
        )
        self.assertEqual(len(data), 1)

    def test_validated_data_assertion_error(self):
        with self.assertRaises(AssertionError):
            self.process_product_batch_requests.validated_data(
                self.not_correct_sample_data
            )

    @patch.object(ProcessProductBatchRequests, 'process_item')
    def test_send(self, mock_process_item):
        mock_process_item.return_value = 'result'
        result = self.process_product_batch_requests.send(self.sample_data)
        self.assertEqual(result, 'result')
        mock_process_item.assert_called_once_with(self.sample_data)

    @patch.object(ProcessProductBatchRequests, 'is_batch_request')
    @patch.object(ProcessProductBatchRequests, 'batch_service')
    def test_check_run_batch_request(
        self,
        mock_batch_service,
        mock_is_batch_request
    ):
        mock_is_batch_request.return_value = True
        self.process_product_batch_requests.integration.batch_request = MagicMock()

        is_ok = False
        formatted_data = 'formatted_data'
        result = self.process_product_batch_requests.check_run(
            is_ok,
            formatted_data
        )

        self.assertFalse(result)
        self.assertIsNone(
            self.process_product_batch_requests.integration.batch_request.objects
        )
        mock_batch_service.assert_called_once_with(self.process_product_batch_requests.integration.channel_id)
        mock_batch_service.return_value.to_fail.assert_called_once_with(
            self.process_product_batch_requests.integration.batch_request
        )

    def test_update_state(self):
        result = self.process_product_batch_requests.update_state
        self.assertEqual(result, BatchRequestStatus.done)

    @patch.object(ProcessProductBatchRequests, 'get_barcode')
    def test_get_channel_items_by_reference_object_ids(self, mock_get_barcode):
        mock_get_barcode.return_value = 'sku1'
        channel_response = [
            MagicMock(sku='sku1'),
            MagicMock(sku='sku2')
        ]
        model_items_by_content = {
            "product": {
                1: Product(id=1, sku='sku1'),
                2: Product(id=2, sku='sku2'),
                3: Product(id=3, sku='sku3')
            }
        }
        integration_actions = []

        result = self.process_product_batch_requests.get_channel_items_by_reference_object_ids(
            channel_response, model_items_by_content, integration_actions
        )

        self.assertEqual(len(result), 3)
        self.assertIn(1, result)
        self.assertIn(2, result)
        self.assertEqual(result[1], channel_response[0])


class TestProcessDeletedProductBatchRequests(BaseTestCaseMixin):
    """
    Test case for ProcessDeletedProductBatchRequests
    run: python -m unittest channel_app.omnitron.commands.tests.test_products.TestProcessDeletedProductBatchRequests
    """

    def setUp(self) -> None:
        self.instance = ProcessDeletedProductBatchRequests(
            integration=self.mock_integration
        )
        self.instance.objects = [
            ProductBatchRequestResponseDto(
                status=ResponseStatus.fail,
                sku='1',
                remote_id=None,
                message=None,
            ),
            ProductBatchRequestResponseDto(
                status=ResponseStatus.success,
                sku='2',
                remote_id=None,
                message=None,
            ),
        ]
        self.sample_response = [
            ProductBatchRequestResponseDto(
                sku='1',
                remote_id=1,
                status=ResponseStatus.success
            ),
            ProductBatchRequestResponseDto(
                sku='2',
                remote_id=2,
                status=ResponseStatus.fail
            )
        ]
        self.integration_actions = [
            MagicMock(content_type={"model": ContentType.product.value}),
            MagicMock(content_type={"model": ContentType.product_price.value}),
            MagicMock(content_type={"model": ContentType.product_stock.value}),
            MagicMock(content_type={"model": ContentType.product_image.value}),
        ]

    def test_get_data(self):
        result = self.instance.get_data()
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].sku, '1')
        self.assertEqual(result[1].sku, '2')

    def test_validated_data(self):
        result = self.instance.validated_data(self.instance.objects)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].sku, '1')
        self.assertEqual(result[1].sku, '2')

    def test_validated_data_but_wrong_data(self):
        with self.assertRaises(AssertionError):
            self.instance.validated_data([{}])
    
    @patch.object(BaseClient, 'get_instance')
    @patch.object(
        ProcessDeletedProductBatchRequests, 
        'get_integration_actions_for_remote_ids'
    )
    @patch.object(
        ChannelIntegrationActionEndpoint,
        '__new__'
    )
    @patch.object(ProcessDeletedProductBatchRequests, 'create_batch_objects')
    @patch.object(ProcessDeletedProductBatchRequests, 'update_batch_request')
    def test_process_item_with_successful_response(
        self,
        mock_update_batch_request,
        mock_create_batch_objects,
        mock_delete,
        mock_get_integration_actions,
        mock_get_instance
    ):
        mock_get_integration_actions.return_value = self.integration_actions

        result = self.instance.process_item(self.sample_response)
        
        self.assertEqual(len(result), 4)
        mock_delete.assert_called_once()

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelIntegrationActionEndpoint, '__new__')
    @patch.object(
        ProcessDeletedProductBatchRequests, 
        'get_integration_actions_for_remote_ids'
    )
    @patch.object(
        ProcessDeletedProductBatchRequests,
        'create_batch_objects'
    )
    @patch.object(
        ProcessDeletedProductBatchRequests,
        'update_batch_request'
    )
    def test_process_item_with_failed_response(
        self,
        mock_update_batch_request,
        mock_create_batch_objects,
        mock_get_integration_actions,
        mock_endpoint,
        mock_get_instance
    ):
        mock_get_integration_actions.return_value = self.integration_actions

        result = self.instance.process_item(self.sample_response)

        self.assertEqual(len(result), 4)
        mock_create_batch_objects.assert_called_with(
            data=self.integration_actions,
            content_type=ContentType.integration_action.value
        )

    @patch.object(
        ProcessDeletedProductBatchRequests, 
        'get_integration_actions_for_remote_ids'
    )
    def test_get_integration_actions_for_remote_ids(
        self,
        mock_get_integration_actions
    ):
        remote_ids = [1, 2, 3, 4, 5]
        mock_get_integration_actions.return_value = [
            MagicMock(remote_id=1),
            MagicMock(remote_id=2),
            MagicMock(remote_id=3),
            MagicMock(remote_id=4),
            MagicMock(remote_id=5),
        ]

        result = self.instance.get_integration_actions_for_remote_ids(remote_ids)

        self.assertEqual(len(result), 5)
        self.assertEqual(result[0].remote_id, 1)
        self.assertEqual(result[1].remote_id, 2)
        self.assertEqual(result[2].remote_id, 3)
        self.assertEqual(result[3].remote_id, 4)
        self.assertEqual(result[4].remote_id, 5)
        


from dataclasses import asdict
from unittest.mock import MagicMock, patch
from omnisdk.base_client import BaseClient
from omnisdk.omnitron.endpoints import (
    ChannelCargoEndpoint, 
    ChannelCustomerEndpoint,
    ChannelOrderEndpoint,
    ChannelOrderItemEndpoint,
    ChannelCancellationRequestEndpoint,
    ChannelBatchRequestEndpoint)
from omnisdk.omnitron.models import (
    CancellationRequest,
    Customer
)

from channel_app.core.data import (
    AddressDto,
    CancellationRequestDto,
    CustomerDto,
    OrderBatchRequestResponseDto,
)
from channel_app.core.tests import BaseTestCaseMixin
from channel_app.omnitron.commands.orders.addresses import GetOrCreateAddress
from channel_app.omnitron.commands.orders.cargo_companies import GetCargoCompany
from channel_app.omnitron.commands.orders.customers import GetOrCreateCustomer
from channel_app.omnitron.commands.orders.orders import (
    CreateCancellationRequest,
    GetCancellationRequestUpdates, 
    GetOrderItems, 
    GetOrderItemsWithOrder, 
    ProcessOrderBatchRequests,
)
from channel_app.omnitron.constants import (
    BatchRequestStatus,
    CancellationType, 
    CustomerIdentifierField
)


class TestProcessOrderBatchRequests(BaseTestCaseMixin):
    """
    Test case for ProcessOrderBatchRequests
    
    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestProcessOrderBatchRequests
    """

    def setUp(self) -> None:
        self.instance = ProcessOrderBatchRequests(
            integration=self.mock_integration,
        )
        self.instance.objects = [
            OrderBatchRequestResponseDto(
                status='fail',
                remote_id='',
                number='1',
                message='Error message'
            ),
            OrderBatchRequestResponseDto(
                status='success',
                remote_id='123',
                number='2',
                message=''
            )
        ]
        self.integration_actions = [
            MagicMock(
                id=1,
                channel=1,
                content_type={
                    'model': 'order'
                },
                remote_id='1',
                object_id=1,
            ),
            MagicMock(
                id=2,
                channel=1,
                content_type={
                    'model': 'order'
                },
                remote_id='2',
                object_id=2,
            )
        ]
        self.order_obj = MagicMock(
            pk=1,
        )

    def test_validated_data(self):
        data = self.instance.validated_data(self.instance.objects)
        self.assertEqual(len(data), 2)
        self.assertIsInstance(data[0], OrderBatchRequestResponseDto)
        self.assertIsInstance(data[1], OrderBatchRequestResponseDto)

    def test_update_state(self):
        result = self.instance.update_state
        self.assertEqual(result, BatchRequestStatus.commit)

    def test_get_remote_order_number(self):
        data = self.instance.get_remote_order_number(
            integration_actions=self.integration_actions,
            obj=self.order_obj,
        )
        self.assertEqual(data, self.integration_actions[0].remote_id)

    def test_get_remote_order_number_content_type_is_not_order(self):
        self.integration_actions[0].content_type = {
            'model': 'product'
        }
        data = self.instance.get_remote_order_number(
            integration_actions=self.integration_actions,
            obj=self.order_obj,
        )
        self.assertIsNone(data)

    def test_get_channel_items_by_reference_object_ids(self):
        channel_response = [
            MagicMock(number='1'),
            MagicMock(number='2'),
        ]
        model_items_by_content = {
            "order": {
                "1": MagicMock(pk=1),
                "2": MagicMock(pk=2),
            }
        }
        integration_actions = self.integration_actions
        result = self.instance.get_channel_items_by_reference_object_ids(
            channel_response,
            model_items_by_content,
            integration_actions
        )
        self.assertEqual(len(result), 2)
        self.assertEqual(result.get('1').number, "1")
        self.assertEqual(result.get('2').number, "2")

    def test_get_orders_with_empty_list(self):
        id_list = []
        result = self.instance.get_orders(id_list)
        self.assertEqual(len(result), 0)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelOrderEndpoint, '_list')
    def test_get_orders(self, mock_endpoint, mock_get_instance):
        response_data = [
            MagicMock(pk=1),
            MagicMock(pk=2),
        ]
        response = MagicMock()
        response.list.return_value = response_data

        with patch.object(ChannelOrderEndpoint, '__new__', return_value=response):
            result = self.instance.get_orders(['1', '2'])
            self.assertEqual(len(result), 2)
            self.assertEqual(result.get(1).pk, response_data[0].pk)
            self.assertEqual(result.get(2).pk, response_data[1].pk)

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ProcessOrderBatchRequests, 'get_orders')
    def test_group_model_items_by_content_type(
        self,
        mock_get_orders, 
        mock_get_instance
    ):
        mock_get_orders.return_value = [
            MagicMock(pk=1),
            MagicMock(pk=2),
        ]
        items_by_content = {
            "order": {
                "1": MagicMock(pk=1),
                "2": MagicMock(pk=2),
            }
        }
        data = self.instance.group_model_items_by_content_type(items_by_content)
        self.assertEqual(len(data), 1)
        self.assertEqual(len(data.get('order')), 2)
        

class TestGetOrCreateCustomer(BaseTestCaseMixin):
    """
    Test case for GetOrCreateCustomer
    
    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestGetOrCreateCustomer
    """

    def setUp(self) -> None:
        self.instance = GetOrCreateCustomer(
            integration=self.mock_integration,
        )
        self.instance.objects: CustomerDto = CustomerDto(
            email="john.doe@akinon.com",
            first_name="John",
            last_name="Doe",
            channel_code="1",
            extra_field={},
            phone_number="05556667788",
            is_active=True,
        )
        self.customer_endpoint_response_data = {
            'pk': 1, 
            'channel': 1, 
            'email': 'john.doe@akinon.com', 
            'first_name': 'John', 
            'last_name': 'Doe', 
            'phone_number': 
            '05556667788', 
            'is_active': True, 
            'channel_code': '1', 
            'erp_code': None, 
            'extra_field': {}, 
            'modified_date': '2024-01-29T08:27:18.399594Z', 
            'created_date': '2024-01-29T08:27:18.399484Z', 
            'date_joined': None, 
            'email_allowed': False, 
            'sms_allowed': False, 
            'call_allowed': False, 
            'gender': None, 
            'attributes': {}, 
            'user_type': None, 
            'date_of_birth': None, 
            'attributes_kwargs': {}, 
            'localized_attributes': {}, 
            'localized_attributes_kwargs': {}
        }
        self.customer_endpoint_response = [
            MagicMock(**self.customer_endpoint_response_data)
        ]

    @patch.object(GetOrCreateCustomer, 'get_customer')
    def test_get_data(self, mock_get_customer):
        mock_get_customer.return_value = [self.instance.objects]
        data = self.instance.get_data()
        
        self.assertEqual(len(data), 1)

        data = data[0]

        for key, value in asdict(self.instance.objects).items():
            self.assertEqual(getattr(data, key), value)

    def test_get_customer_identifier_is_email(self):
        self.instance.integration.channel.conf = {
            "CUSTOMER_IDENTIFIER_FIELD": CustomerIdentifierField.email
        }

        response = MagicMock()
        response.list.return_value = self.customer_endpoint_response

        with patch.object(
            ChannelCustomerEndpoint, 
            '__new__', 
            return_value=response
        ):
            customers = self.instance.get_customer(self.instance.objects)
            self.assertEqual(len(customers), 1)

            customer = customers[0]

            for key, value in self.customer_endpoint_response_data.items():
                self.assertEqual(getattr(customer, key), value)
            

    def test_get_customer_identifier_is_phone_number(self):
        self.instance.integration.channel.conf = {
            "CUSTOMER_IDENTIFIER_FIELD": CustomerIdentifierField.phone_number
        }

        response = MagicMock()
        response.list.return_value = self.customer_endpoint_response

        with patch.object(
            ChannelCustomerEndpoint,
            '__new__',
            return_value=response
        ):
            customers = self.instance.get_customer(self.instance.objects)
            self.assertEqual(len(customers), 1)

            customer = customers[0]

            for key, value in self.customer_endpoint_response_data.items():
                self.assertEqual(getattr(customer, key), value)
            
    def test_get_customer_email_filter_incorrect_exception(self):
        self.instance.integration.channel.conf = {
            "CUSTOMER_IDENTIFIER_FIELD": CustomerIdentifierField.email
        }

        response = MagicMock()
        response.list.return_value = [MagicMock()]

        with patch.object(
            ChannelCustomerEndpoint,
            '__new__',
            return_value=response
        ):
            with self.assertRaises(
                Exception, 
                msg="Customer email filter incorrect"
            ):
                self.instance.get_customer(self.instance.objects)

    def test_get_customer_phone_number_filter_incorrect_exception(self):
        self.instance.integration.channel.conf = {
            "CUSTOMER_IDENTIFIER_FIELD": CustomerIdentifierField.phone_number
        }

        response = MagicMock()
        response.list.return_value = [MagicMock()]

        with patch.object(
            ChannelCustomerEndpoint,
            '__new__',
            return_value=response
        ):
            with self.assertRaises(
                Exception, 
                msg="Customer phone_number filter incorrect"
            ):
                self.instance.get_customer(self.instance.objects)

    def test_get_customer_must_update_case(self):
        self.instance.integration.channel.conf = {
            "CUSTOMER_IDENTIFIER_FIELD": CustomerIdentifierField.email
        }

        self.instance.objects.first_name = "Jenny"
        self.instance.objects.last_name = "Doey"
        
        response = MagicMock()
        response.list.return_value = self.customer_endpoint_response
        response.update.return_value = MagicMock(
            **self.customer_endpoint_response_data
        )

        with patch.object(
            ChannelCustomerEndpoint,
            '__new__',
            return_value=response
        ):
            customers = self.instance.get_customer(self.instance.objects)
            self.assertEqual(len(customers), 1)

            customer = customers[0]

            for key, value in self.customer_endpoint_response_data.items():
                self.assertEqual(getattr(customer, key), value)


class TestGetOrCreateAddress(BaseTestCaseMixin):
    """
    Test case for GetOrCreateAddress

    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestGetOrCreateAddress
    """

    def setUp(self) -> None:
        self.instance = GetOrCreateAddress(
            integration=self.mock_integration,
        )
        self.instance.objects = {}
        self.instance.objects["address"]: AddressDto = AddressDto(
            email="john.doe@akinon.com",
            phone_number="05556667788",
            first_name="John",
            last_name="Doe",
            country="1",
            city="80",
            line="Hemen sahil kenarı"
        )
        self.instance.objects["customer"]: Customer = Customer(
            pk="1",
            first_name="Customer John",
            last_name="Customer Doe",
        )


    @patch.object(GetOrCreateAddress, 'get_location_objects')
    def test_get_data(self, mock_get_location_objects):
        mock_get_location_objects.return_value = (1, 80, None, None)
        data = self.instance.get_data()

        self.assertEqual(
            data["first_name"],
            self.instance.objects["address"].first_name
        )
        self.assertEqual(
            data["last_name"],
            self.instance.objects["address"].last_name
        )


class TestGetCargoCompany(BaseTestCaseMixin):
    """
    Test case for GetCargoCompany

    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestGetCargoCompany
    """

    def setUp(self) -> None:
        self.instance = GetCargoCompany(
            integration=self.mock_integration,
        )
        self.instance.objects = 'zyWpJvJdlqRaUfZMtdxgEnzslqXobZWxDGjtlQZtwQUnHtGkpJYzzJuoJhSVQIUW'
        self.channel_cargo_endpoint_response_data = {
            'pk': 1, 
            'name': 
            'ltLIdxIpZlyksSTPEGBytajaRazeShlCzrUZQBIteaHxQElXzQglgnZHZUyuzSqH', 
            'erp_code': 'zyWpJvJdlqRaUfZMtdxgEnzslqXobZWxDGjtlQZtwQUnHtGkpJYzzJuoJhSVQIUW', 
            'shipping_company': 'bringo_express', 
            'modified_date': '2024-01-29T10:04:56.791090Z', 
            'created_date': '2024-01-29T10:04:56.790945Z'
        }
        self.channel_cargo_endpoint_response = [
            MagicMock(**self.channel_cargo_endpoint_response_data)
        ]

    def test_get_cargo_company(self):
        cargo_company = self.instance.get_cargo_company(
            self.channel_cargo_endpoint_response
        )
        self.assertEqual(cargo_company.erp_code, self.instance.objects)

    def test_get_cargo_company_not_exists_exception(self):
        self.instance.objects = 'not_exists_erp_code'

        with self.assertRaises(
            Exception, 
            msg=f"CargoCompany does not exists: {self.instance.objects}"
        ):
            self.instance.get_cargo_company(
                self.channel_cargo_endpoint_response
            )

    @patch.object(GetCargoCompany, 'get_cargo_company')
    def test_get_data(self, mock_get_cargo_company):
        mock_get_cargo_company.return_value = self.channel_cargo_endpoint_response[0]

        response = MagicMock()
        response.list.return_value = self.channel_cargo_endpoint_response
        response.iterator = iter(self.channel_cargo_endpoint_response)

        with patch.object(
            ChannelCargoEndpoint,
            '__new__',
            return_value=response
        ):
            cargo_companies = self.instance.get_data()
            self.assertEqual(len(cargo_companies), 1)

            cargo_company = cargo_companies[0]

            self.assertEqual(
                cargo_company.erp_code, 
                self.channel_cargo_endpoint_response_data['erp_code']
            )
        

class TestGetOrderItems(BaseTestCaseMixin):
    """
    Test case for GetOrderItems
    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestGetOrderItems
    """
    def setUp(self) -> None:
        self.instance = GetOrderItems(
            integration=self.mock_integration,
        )
        self.order = MagicMock(
            pk=1,
        )
        self.order_items = [
            MagicMock(
                pk=7,
                order=1,
                content_type='orderitem'
            ),
            MagicMock(
                pk=8,
                order=1,
                content_type='orderitem'
            )
        ]
        self.instance.objects = self.order

    @patch.object(GetOrderItems, 'get_order_items')
    def test_get_data(self, mock_get_order_items):
        mock_get_order_items.return_value = self.order_items
        order_items = self.instance.get_data()
        
        self.assertEqual(len(order_items), 2)

        for order_item in order_items:
            self.assertEqual(order_item.order, self.order.pk)
            self.assertEqual(order_item.content_type, 'orderitem')

    def test_get_order_items(self):
        response = MagicMock()
        response.list.return_value = self.order_items

        with patch.object(
            ChannelOrderItemEndpoint,
            '__new__',
            return_value=response
        ):
            order_items = self.instance.get_order_items(self.order)
            self.assertEqual(len(order_items), 2)

            for order_item in order_items:
                self.assertEqual(order_item.order, self.order.pk)
                self.assertEqual(order_item.content_type, 'orderitem')


class TestGetOrderItemsWithOrder(BaseTestCaseMixin):
    """
    Test case for GetOrderItems
    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestGetOrderItemsWithOrder
    """
    def setUp(self) -> None:
        self.instance = GetOrderItemsWithOrder(
            integration=self.mock_integration,
        )
        self.orders = [
            MagicMock(
                pk=1,
                orderitem_set=[],
            ),
        ]
        self.order_items = [
            MagicMock(
                pk=7,
                order=1,
                content_type='orderitem'
            ),
            MagicMock(
                pk=8,
                order=1,
                content_type='orderitem'
            ),
        ]
        self.instance.objects = self.orders

    @patch.object(GetOrderItemsWithOrder, 'get_order_items')
    def test_get_data(self, mock_get_order_items):
        mock_get_order_items.return_value = self.order_items
        orders = self.instance.get_data()

        self.assertEqual(len(orders), 1)
        self.assertEqual(len(orders[0].orderitem_set), 2)

class TestGetCancellationRequestUpdates(BaseTestCaseMixin):
    """
    Test case for GetCancellationRequestUpdates
    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.GetCancellationRequestUpdates
    """

    def setUp(self) -> None:
        self.instance = GetCancellationRequestUpdates(
            integration=self.mock_integration,
        )
       
        self.instance.objects = {
            'status': 'completed',
            'cancellation_type': 'refund'
        }
        self.endpoint = ChannelCancellationRequestEndpoint
        self.cancellation_request = MagicMock(
                pk=1,
                status='completed',
                cancellation_type='refund',
                easy_return=None,
                order_item=1,
            )
        return super().setUp()

    @patch.object(BaseClient, 'get_instance')
    @patch.object(ChannelCancellationRequestEndpoint, '_list')
    @patch.object(ChannelBatchRequestEndpoint, 'update')    
    def test_get_data(self, mock_endpoint, mock_get_instance, mock_batch_request):
        response_data = [
                MagicMock(
                pk=1,
                status='completed',
                cancellation_type=CancellationType.refund.value,
                easy_return=None,
                order_item=1,
            ),
                MagicMock(
                pk=2,
                status='completed',
                cancellation_type=CancellationType.refund.value,
                easy_return=None,
                order_item=2,
            )
        ]
        response = MagicMock()
        response.list.return_value = response_data
        with patch.object(self.endpoint, '__new__', return_value=response):
            data = self.instance.get_data()
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0].status, self.instance.objects['status'])

    @patch.object(GetCancellationRequestUpdates, 'get_cancellation_requests')
    def test_update_cancellation_request(self, mock_get_cancellation_requests):
        mock_get_cancellation_requests.return_value = self.cancellation_request
        self.instance.get_cancellation_requests(self.cancellation_request)
        self.assertEqual(self.cancellation_request.status, self.instance.objects['status'])
        self.assertEqual(self.cancellation_request.cancellation_type, self.instance.objects['cancellation_type'])


class TestCreateCancellationRequest(BaseTestCaseMixin):
    """
    Test case for CreateCancellationRequest
    run: python -m unittest channel_app.omnitron.commands.tests.test_orders.TestCreateCancellationRequest
    """

    def setUp(self) -> None:
        self.instance = CreateCancellationRequest(
            integration=self.mock_integration)

        self.endpoint = ChannelCancellationRequestEndpoint
        self.instance.objects = CancellationRequestDto(
            order_item='1',
            reason='reason_code',
            remote_id='2',
            cancellation_type=CancellationType.refund.value,
            description='description',
        )
        return super().setUp()

    @patch.object(BaseClient, 'get_instance')
    @patch.object(CreateCancellationRequest, 'get_omnitron_order_item')
    def test_get_data(self, mock_get_instance, mock_get_omnitron_order_item):
        mock_get_omnitron_order_item.return_value = 1
        response_data = CancellationRequest(
                pk=1,
                status='waiting',
                cancellation_type='refund',
                easy_return=None,
                order_item=1,
            )
        response = MagicMock()
        response.create.return_value = response_data

        with patch.object(self.endpoint, '__new__', return_value=response):
            data = self.instance.get_data()
            self.assertIsInstance(data, CancellationRequest)



import unittest
from unittest.mock import MagicMock
from channel_app.core.commands import OmnitronCommandInterface


class BaseTestCaseMixin(unittest.TestCase):
    mock_integration = MagicMock()
    omnitron_command_interface = OmnitronCommandInterface(
        integration=mock_integration
    )
    
    def setUp(self) -> None:
        return super().setUp()
    
    def tearDown(self) -> None:
        return super().tearDown()
